/* Generated By:JavaCC: Do not edit this line. IntProlog.java */
package aplprolog.Parsing;
import java.io.StringReader;
import java.util.HashMap;
import aplprolog.datastructures.PrimIntArrayList;
import aplprolog.datastructures.PrimIntKeyHashMap;
import aplprolog.prolog.IntHarvester;
import aplprolog.prolog.StringStorage;
/**
 * The parser to convert <code>String</code> objects to integer arrays. Viewing the "IntProlog.jj" file is advised. 
 * 
 * @author Bas Testerink, Utrecht University, The Netherlands
 * 
 */
public class IntProlog implements IntPrologConstants {
        // rules: parsed rules, splits: split points of subgoals in current query/rule
        // varIndices: indices inside the current array where variable id's are
        // vars: mapping from string var names in the query to integer id's
        // name_arity & nextIsHead: bookkeeping for adding rule
        // strStor: mapping from String id to String and back
        public PrimIntKeyHashMap<PrimIntKeyHashMap<PrimIntArrayList>> rules = new PrimIntKeyHashMap<PrimIntKeyHashMap<PrimIntArrayList>>(PrimIntKeyHashMap.OBJ);
        public PrimIntArrayList array = new PrimIntArrayList(true), splits = new PrimIntArrayList(true);
        public HashMap<Integer,PrimIntArrayList> varIndices = new HashMap<Integer, PrimIntArrayList>();
        public HashMap<String,Integer> vars = new HashMap<String,Integer>();
        public int[] name_arity = new int[]{0,0};
        public boolean nextIsHead = false;
        public StringStorage strStor;

 /** Main entry point. Contains now solely some testing code. */
  public static void main(String args[]) throws ParseException {
    IntProlog parser = new IntProlog(System.in);
        StringStorage strStor = new StringStorage();
        parser.strStor = strStor;
        //parser.addBuiltInStrings();
    parser.parseRulebase(strStor);
        //System.out.println(parser.arStr(parser.parseFact("a(A).",strStor)));
        //System.out.println(parser.arStr(parser.parseQuery("A is -2.",strStor)));
  }

  /**
  * Parse a collection of facts/rules.
  * @param strStor String mapping.
  */
  public void parseRulebase(StringStorage strStor){
    this.strStor = strStor;
        try{
                Rules();
        }catch(ParseException e){System.out.println(e);}
  }

  /**
  * Parse a query. Resulting array will always have a 0 at the beginning of the splits (required for the resolution mechanism).
  * @param strStor String mapping.
  */
  public int[] parseQuery(String q, StringStorage strStor) throws ParseException {
        this.strStor = strStor;
        ReInit(new StringReader("true :- "+q)); // add dummy head
        Rule(); // parse the query
        for(int var = vars.size()-1; var >= 0; var--){ // add overhead variable integers
                PrimIntArrayList indices = varIndices.get(var);
                for(int i = indices.pointer-1; i >= 0; i--) array.addInt(indices.intdata[i]-3); // remove the length of "true" from the indices
                array.addInt(indices.pointer);
        }
        array.addInt(vars.size());
        for(int split = splits.pointer-1; split >= 0; split--) // add overhead split integers 
                array.addInt(splits.intdata[split]-3); // remove the length of "true" from the indices
        array.addInt(splits.pointer);
        return getPrimArray(3);
  }

  /**
  * Parse a term/fact.
  * @param f Fact to be parsed.
  * @param strStor String mapping.
  */
  public int[] parseFact(String f, StringStorage strStor, boolean withOverhead){
        try{
                this.strStor = strStor;
                ReInit(new StringReader(f));
                Rule(); // parse it
        if(withOverhead){
                addVarOverhead(); // add overhead integers
            addSplitsOverhead();
        }
                return getPrimArray(0); // return result
        }catch(ParseException e){System.out.println(e);}
        return null;
  }

  public int[] parseTerm(String f, StringStorage strStor){
        try{
                this.strStor = strStor;
                ReInit(new StringReader(f));
                makeReadyForNextRule();
                Argument(); // parse it
                return getPrimArray(0); // return result
        }catch(ParseException e){System.out.println(e);}
        return null;
  }

  /**
  * Pure for testing, real built-in strings and evaluation happens in the built-in class and aplprolog.prolog.BuiltIn.java.
  */
  public void addBuiltInStrings(){
        String[] intercepts = new String[]{"=","\u005c\u005c=","var","atom","integer","float","atomic","compound","nonvar","number","@=<","==","\u005c\u005c==","@<","@>","@>=","functor","arg","=..","copy_term"," is ","=:=","=\u005c\u005c=","<","=<",">",">=","clause","current_predicate","asserta","assertz","retract","abolish","findall","bagof","setof","once","repeat","atom_length","atom_concat","sub_atom","atom_chars","atom_codes","char_code","number_chars","number_codes","random","true","int_to_atom","fail"};
        String[] functions = new String[]{"+","-","*","/"," div "," rem "," mod ","abs","sign","float_integer_part","float_fractional_part","float","floor","truncate","round","**","^","sin","cos","atan","exp","log","sqrt",">>","<<","/\u005c\u005c","\u005c\u005c/","\u005c\u005c","max","min","integer"," xor ","tan","asin","acos","pi","e"};
        for(String s : intercepts)strStor.add(s);
        for(String s : functions)strStor.add(s);
  }

  /**
  * Clear out the rule-related data.
  */
  public void makeReadyForNextRule(){
        array.clear();
        vars.clear();
        varIndices.clear();
        splits.clear();
  }

  /**
  * Clear out rule base related data.
  */
  public void makeReadyForNextRuleBase(){
        rules.clear();
        nextIsHead = true;
  }

  /**
  * Get the id of a string, given the current string mapping. If it is not encountered yet,
  * then add it to the string mapping.
  * @param s String to get the id from.
  */
  public int strID(String s){
        Integer id = strStor.getInt(s);
        if(id==null) id = strStor.add(s);
        return id;
  }

  /**
  * Get the integer representation of the variable with String identifier v. Add it
  * to the variable map if needed, and keep track of where it was found in the array.
  * @param v String name of the variable.
  */
  public int handleVar(String v){
                if(v.equals("_"))v+="G"+vars.size(); // anonymous variables share names in notation
                if(vars.get(v)==null) vars.put(v,vars.size()); // add it if needed
                int id = vars.get(v);
                if(varIndices.get(id)==null)varIndices.put(id,new PrimIntArrayList(true));
                varIndices.get(id).addInt(array.pointer); // remember where the variable was found
                return id;
        }

  /**
  * Adds the current array as a rule to the rule base.
  */
  public void addRule(){
        // space in rule base
        PrimIntKeyHashMap<PrimIntArrayList> name = rules.getObj(name_arity[0]);
        if(name==null){ name = new PrimIntKeyHashMap<PrimIntArrayList>(PrimIntKeyHashMap.OBJ); rules.putObj(name_arity[0],name); }
        PrimIntArrayList arity = name.getObj(name_arity[1]);
        if(arity==null){ arity = new PrimIntArrayList(false); name.putObj(name_arity[1],arity);}

        // splits and var indices (overhead integers)
        addVarOverhead();
        addSplitsOverhead();

        // conversion arraylist -> primitive array
        arity.addAr(getPrimArray(0));
        //System.out.println(strStor.getString(name_arity[0]) + "/"+name_arity[1]+": "+array);
  }

  /**
  * Adds integers about variable positions in the array.
  */
  public void addVarOverhead(){
        for(int var = vars.size()-1; var >= 0; var--){
                PrimIntArrayList indices = varIndices.get(var);
                for(int i = indices.pointer-1; i >= 0; i--) array.addInt(indices.intdata[i]);
                array.addInt(indices.pointer);
        }
        array.addInt(vars.size());
  }

  /**
  * Adds integers about split positions in the array.
  */
  public void addSplitsOverhead(){
        for(int split = splits.pointer-1; split >= 0; split--)
                array.addInt(splits.intdata[split]);
        array.addInt(splits.pointer);
  }

  /**
  * Convert the working arraylist to a primitive array.
  */
  public int[] getPrimArray(int start){
        int[] primArray = new int[array.pointer-start];
        for(int i = 0; i < primArray.length; i++)
                primArray[i] = array.intdata[i+start];
        return primArray;
  }

  /**
  * Output function: convert int[] to String.
  */
  public String arStr(int[] ar){
                if(ar==null) return "NULL";
                String s = "[";
                for(int i = 0; i < ar.length; i++){
                        s+=ar[i];
                        if(i!=ar.length-1)s+=",";
                }
                return s + "]";
  }

  final public void Rules() throws ParseException {
 makeReadyForNextRuleBase();
    label_1:
    while (true) {
      switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
      case CHARACTER_LITERAL:
      case STD_IDENTIFIER:
      case SEMICOLON:
      case DOT:
      case IMPLY:
      case RULEIMPLY:
      case BAR:
        ;
        break;
      default:
        jj_la1[0] = jj_gen;
        break label_1;
      }
      Rule();
                        nextIsHead = true;
                        addRule();
    }
    jj_consume_token(0);
  }

  final public void Rule() throws ParseException {
  makeReadyForNextRule();
    switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
    case CHARACTER_LITERAL:
    case STD_IDENTIFIER:
    case SEMICOLON:
    case IMPLY:
    case BAR:
      PosLiteral();
                       splits.addInt(array.pointer);
      break;
    default:
      jj_la1[1] = jj_gen;
      ;
    }
    switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
    case RULEIMPLY:
      jj_consume_token(RULEIMPLY);
      Argument();
                                        splits.addInt(array.pointer);
      label_2:
      while (true) {
        switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
        case COMMA:
          ;
          break;
        default:
          jj_la1[2] = jj_gen;
          break label_2;
        }
        jj_consume_token(COMMA);
        Argument();
                                     splits.addInt(array.pointer);
      }
      break;
    default:
      jj_la1[3] = jj_gen;
      ;
    }
    jj_consume_token(DOT);
  }

  final public void PosLiteral() throws ParseException {
  Token t; int currentSize = 0; int nrArgs = 0; boolean thisIsHead = nextIsHead; nextIsHead = false;
    switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
    case STD_IDENTIFIER:
      t = jj_consume_token(STD_IDENTIFIER);
      break;
    case CHARACTER_LITERAL:
      t = jj_consume_token(CHARACTER_LITERAL);
      break;
    case IMPLY:
      t = jj_consume_token(IMPLY);
      break;
    case SEMICOLON:
      t = jj_consume_token(SEMICOLON);
      break;
    case BAR:
      t = jj_consume_token(BAR);
      break;
    default:
      jj_la1[4] = jj_gen;
      jj_consume_token(-1);
      throw new ParseException();
    }
                        array.addInt(IntHarvester.PREDICATE);
                        array.addInt(strID(t.image));
                        currentSize = array.pointer;
                        array.addInt(0);
                        if(thisIsHead){
                                name_arity[0] = strID(t.image);
                                name_arity[1] = 0;
                        }
    switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
    case LPAREN:
      nrArgs = Arguments();
                        array.intdata[currentSize] = nrArgs;
                        if(thisIsHead)name_arity[1] = nrArgs;
      break;
    default:
      jj_la1[5] = jj_gen;
      ;
    }
  }

  final public void NegLiteral() throws ParseException {
    jj_consume_token(NOT);
                array.addInt(IntHarvester.PREDICATE);
                array.addInt(strID("not"));
                array.addInt(1);
    if (jj_2_1(2147483647)) {
      Expression();
    } else {
      switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
      case CHARACTER_LITERAL:
      case STD_IDENTIFIER:
      case SEMICOLON:
      case IMPLY:
      case BAR:
        PosLiteral();
        break;
      case LPAREN:
        Parenthesized();
        break;
      case VAR_IDENTIFIER:
        Variable();
        break;
      default:
        jj_la1[6] = jj_gen;
        jj_consume_token(-1);
        throw new ParseException();
      }
    }
  }

  final public int Arguments() throws ParseException {
 int counter = 1;
    jj_consume_token(LPAREN);
    Argument();
    label_3:
    while (true) {
      switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
      case COMMA:
        ;
        break;
      default:
        jj_la1[7] = jj_gen;
        break label_3;
      }
      jj_consume_token(COMMA);
      Argument();
                                                 counter++;
    }
    jj_consume_token(RPAREN);
         {if (true) return counter;}
    throw new Error("Missing return statement in function");
  }

  final public void Argument() throws ParseException {
    if (jj_2_2(2147483647)) {
      Expression();
    } else {
      switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
      case SUBTRACT:
        NegativeNumber();
        break;
      case CHARACTER_LITERAL:
      case STD_IDENTIFIER:
      case SEMICOLON:
      case IMPLY:
      case BAR:
        PosLiteral();
        break;
      case NOT:
        NegLiteral();
        break;
      case INTEGER_LITERAL:
      case FLOATING_POINT_LITERAL:
        Number();
        break;
      case VAR_IDENTIFIER:
        Variable();
        break;
      case BANG:
        Cut();
        break;
      case LPAREN:
        Parenthesized();
        break;
      case LBRACKET:
        List();
        break;
      default:
        jj_la1[8] = jj_gen;
        jj_consume_token(-1);
        throw new ParseException();
      }
    }
  }

  final public void Parenthesized() throws ParseException {
 int currentSize = 0; int nrArgs = 0;
                array.addInt(IntHarvester.PARENTHESIZED);
                currentSize = array.pointer;
                array.addInt(1);
    nrArgs = Arguments();
                array.intdata[currentSize] = nrArgs;
  }

  final public void List() throws ParseException {
 int counter = 0; int currentSize = 0; int nrArgs = 0; boolean hasTail = false;
                array.addInt(IntHarvester.LIST);
                currentSize = array.pointer;
                array.addInt(0);
    jj_consume_token(LBRACKET);
    switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
    case INTEGER_LITERAL:
    case FLOATING_POINT_LITERAL:
    case CHARACTER_LITERAL:
    case STD_IDENTIFIER:
    case VAR_IDENTIFIER:
    case LPAREN:
    case LBRACKET:
    case SEMICOLON:
    case BANG:
    case SUBTRACT:
    case IMPLY:
    case NOT:
    case BAR:
      Argument();
                              nrArgs++;
      label_4:
      while (true) {
        switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
        case COMMA:
          ;
          break;
        default:
          jj_la1[9] = jj_gen;
          break label_4;
        }
        jj_consume_token(COMMA);
        Argument();
                                                              nrArgs++;
      }
      switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
      case BAR:
        jj_consume_token(BAR);
        switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
        case LBRACKET:
          List();
          break;
        case VAR_IDENTIFIER:
          Variable();
          break;
        default:
          jj_la1[10] = jj_gen;
          jj_consume_token(-1);
          throw new ParseException();
        }
                                            hasTail = true;
        break;
      default:
        jj_la1[11] = jj_gen;
        ;
      }
      break;
    default:
      jj_la1[12] = jj_gen;
      ;
    }
                                array.intdata[currentSize] = nrArgs;
                                if(nrArgs>0&&!hasTail){
                                        array.addInt(IntHarvester.LIST);
                                        array.addInt(0);
                                }
    jj_consume_token(RBRACKET);
  }

  final public void Number() throws ParseException {
 Token t;
    switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
    case INTEGER_LITERAL:
      t = jj_consume_token(INTEGER_LITERAL);
      break;
    case FLOATING_POINT_LITERAL:
      t = jj_consume_token(FLOATING_POINT_LITERAL);
      break;
    default:
      jj_la1[13] = jj_gen;
      jj_consume_token(-1);
      throw new ParseException();
    }
                array.addInt(IntHarvester.NUMBER);
                double d = Double.parseDouble(t.image);
                long l = Double.doubleToLongBits(d);
                int a = (int)(l>>>32);
                int b = (int)((l<<32)>>>32);
                array.addInt(a);
                array.addInt(b);
  }

  final public void Variable() throws ParseException {
 Token t;
    t = jj_consume_token(VAR_IDENTIFIER);
                array.addInt(IntHarvester.VARIABLE);
                array.addInt(handleVar(t.image));
  }

  final public void Cut() throws ParseException {
    jj_consume_token(BANG);
                array.addInt(IntHarvester.CUT);
                array.addInt(0);
  }

  final public Integer Operator() throws ParseException {
 Token t;
    switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
    case ADD:
      t = jj_consume_token(ADD);
      break;
    case SUBTRACT:
      t = jj_consume_token(SUBTRACT);
      break;
    case IMPLY:
      t = jj_consume_token(IMPLY);
      break;
    case MULTIPLY:
      t = jj_consume_token(MULTIPLY);
      break;
    case POWER:
      t = jj_consume_token(POWER);
      break;
    case DIVIDE:
      t = jj_consume_token(DIVIDE);
      break;
    case BITAND:
      t = jj_consume_token(BITAND);
      break;
    case UNIFICATION:
      t = jj_consume_token(UNIFICATION);
      break;
    case ARITHUNIF:
      t = jj_consume_token(ARITHUNIF);
      break;
    case ARITHNOTUNIF:
      t = jj_consume_token(ARITHNOTUNIF);
      break;
    case EQLT:
      t = jj_consume_token(EQLT);
      break;
    case TERMDECOMPOSITION:
      t = jj_consume_token(TERMDECOMPOSITION);
      break;
    case EQUALS:
      t = jj_consume_token(EQUALS);
      break;
    case BACKSLASH:
      t = jj_consume_token(BACKSLASH);
      break;
    case BITOR:
      t = jj_consume_token(BITOR);
      break;
    case NOTUNIFIABLE:
      t = jj_consume_token(NOTUNIFIABLE);
      break;
    case NOTEQUALS:
      t = jj_consume_token(NOTEQUALS);
      break;
    case TERMGTEQ:
      t = jj_consume_token(TERMGTEQ);
      break;
    case TERMEQLT:
      t = jj_consume_token(TERMEQLT);
      break;
    case TERMLT:
      t = jj_consume_token(TERMLT);
      break;
    case TERMGT:
      t = jj_consume_token(TERMGT);
      break;
    case TERMEQ:
      t = jj_consume_token(TERMEQ);
      break;
    case TERMNEQ:
      t = jj_consume_token(TERMNEQ);
      break;
    case LT:
      t = jj_consume_token(LT);
      break;
    case SHIFTLEFT:
      t = jj_consume_token(SHIFTLEFT);
      break;
    case GT:
      t = jj_consume_token(GT);
      break;
    case SHIFTRIGHT:
      t = jj_consume_token(SHIFTRIGHT);
      break;
    case GTEQ:
      t = jj_consume_token(GTEQ);
      break;
    case ASSIGN:
      t = jj_consume_token(ASSIGN);
      break;
    case POWER2:
      t = jj_consume_token(POWER2);
      break;
    case XOR:
      t = jj_consume_token(XOR);
      break;
    case REMAINDER:
      t = jj_consume_token(REMAINDER);
      break;
    case INTDIVIDE:
      t = jj_consume_token(INTDIVIDE);
      break;
    case MODULO:
      t = jj_consume_token(MODULO);
      break;
    case COLON:
      t = jj_consume_token(COLON);
      break;
    case RULEIMPLY:
      t = jj_consume_token(RULEIMPLY);
      break;
    case SEMICOLON:
      t = jj_consume_token(SEMICOLON);
      break;
    default:
      jj_la1[14] = jj_gen;
      jj_consume_token(-1);
      throw new ParseException();
    }
         {if (true) return strID(t.image);}
    throw new Error("Missing return statement in function");
  }

  final public void Expression() throws ParseException {
 Integer opid; int currentSize = 0;
                array.addInt(IntHarvester.PREDICATE);
                currentSize = array.pointer;
                array.addInt(-1);
                array.addInt(2);
    switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
    case SUBTRACT:
      NegativeNumber();
      break;
    case CHARACTER_LITERAL:
    case STD_IDENTIFIER:
    case SEMICOLON:
    case IMPLY:
    case BAR:
      PosLiteral();
      break;
    case NOT:
      NegLiteral();
      break;
    case INTEGER_LITERAL:
    case FLOATING_POINT_LITERAL:
      Number();
      break;
    case VAR_IDENTIFIER:
      Variable();
      break;
    case BANG:
      Cut();
      break;
    case LPAREN:
      Parenthesized();
      break;
    case LBRACKET:
      List();
      break;
    default:
      jj_la1[15] = jj_gen;
      jj_consume_token(-1);
      throw new ParseException();
    }
    opid = Operator();
    Argument();
         array.intdata[currentSize] = opid;
  }

  final public void NegativeNumber() throws ParseException {
                array.addInt(IntHarvester.PREDICATE);
                array.addInt(strID("-"));
                array.addInt(1);
    jj_consume_token(SUBTRACT);
    Argument();
  }

  private boolean jj_2_1(int xla) {
    jj_la = xla; jj_lastpos = jj_scanpos = token;
    try { return !jj_3_1(); }
    catch(LookaheadSuccess ls) { return true; }
    finally { jj_save(0, xla); }
  }

  private boolean jj_2_2(int xla) {
    jj_la = xla; jj_lastpos = jj_scanpos = token;
    try { return !jj_3_2(); }
    catch(LookaheadSuccess ls) { return true; }
    finally { jj_save(1, xla); }
  }

  private boolean jj_3R_22() {
    if (jj_3R_38()) return true;
    return false;
  }

  private boolean jj_3R_16() {
    if (jj_scan_token(SUBTRACT)) return true;
    if (jj_3R_15()) return true;
    return false;
  }

  private boolean jj_3R_30() {
    if (jj_3R_21()) return true;
    return false;
  }

  private boolean jj_3R_15() {
    Token xsp;
    xsp = jj_scanpos;
    if (jj_3R_24()) {
    jj_scanpos = xsp;
    if (jj_3R_25()) {
    jj_scanpos = xsp;
    if (jj_3R_26()) {
    jj_scanpos = xsp;
    if (jj_3R_27()) {
    jj_scanpos = xsp;
    if (jj_3R_28()) {
    jj_scanpos = xsp;
    if (jj_3R_29()) {
    jj_scanpos = xsp;
    if (jj_3R_30()) {
    jj_scanpos = xsp;
    if (jj_3R_31()) {
    jj_scanpos = xsp;
    if (jj_3R_32()) return true;
    }
    }
    }
    }
    }
    }
    }
    }
    return false;
  }

  private boolean jj_3R_24() {
    if (jj_3R_5()) return true;
    return false;
  }

  private boolean jj_3_1() {
    if (jj_3R_5()) return true;
    return false;
  }

  private boolean jj_3R_6() {
    if (jj_3R_16()) return true;
    return false;
  }

  private boolean jj_3R_38() {
    if (jj_scan_token(LPAREN)) return true;
    if (jj_3R_15()) return true;
    Token xsp;
    while (true) {
      xsp = jj_scanpos;
      if (jj_3R_40()) { jj_scanpos = xsp; break; }
    }
    if (jj_scan_token(RPAREN)) return true;
    return false;
  }

  private boolean jj_3R_29() {
    if (jj_3R_20()) return true;
    return false;
  }

  private boolean jj_3R_5() {
    Token xsp;
    xsp = jj_scanpos;
    if (jj_3R_6()) {
    jj_scanpos = xsp;
    if (jj_3R_7()) {
    jj_scanpos = xsp;
    if (jj_3R_8()) {
    jj_scanpos = xsp;
    if (jj_3R_9()) {
    jj_scanpos = xsp;
    if (jj_3R_10()) {
    jj_scanpos = xsp;
    if (jj_3R_11()) {
    jj_scanpos = xsp;
    if (jj_3R_12()) {
    jj_scanpos = xsp;
    if (jj_3R_13()) return true;
    }
    }
    }
    }
    }
    }
    }
    if (jj_3R_14()) return true;
    if (jj_3R_15()) return true;
    return false;
  }

  private boolean jj_3R_34() {
    if (jj_3R_5()) return true;
    return false;
  }

  private boolean jj_3R_13() {
    if (jj_3R_23()) return true;
    return false;
  }

  private boolean jj_3R_28() {
    if (jj_3R_19()) return true;
    return false;
  }

  private boolean jj_3R_18() {
    if (jj_scan_token(NOT)) return true;
    Token xsp;
    xsp = jj_scanpos;
    if (jj_3R_34()) {
    jj_scanpos = xsp;
    if (jj_3R_35()) {
    jj_scanpos = xsp;
    if (jj_3R_36()) {
    jj_scanpos = xsp;
    if (jj_3R_37()) return true;
    }
    }
    }
    return false;
  }

  private boolean jj_3R_14() {
    Token xsp;
    xsp = jj_scanpos;
    if (jj_scan_token(35)) {
    jj_scanpos = xsp;
    if (jj_scan_token(36)) {
    jj_scanpos = xsp;
    if (jj_scan_token(37)) {
    jj_scanpos = xsp;
    if (jj_scan_token(38)) {
    jj_scanpos = xsp;
    if (jj_scan_token(39)) {
    jj_scanpos = xsp;
    if (jj_scan_token(40)) {
    jj_scanpos = xsp;
    if (jj_scan_token(41)) {
    jj_scanpos = xsp;
    if (jj_scan_token(42)) {
    jj_scanpos = xsp;
    if (jj_scan_token(43)) {
    jj_scanpos = xsp;
    if (jj_scan_token(44)) {
    jj_scanpos = xsp;
    if (jj_scan_token(45)) {
    jj_scanpos = xsp;
    if (jj_scan_token(46)) {
    jj_scanpos = xsp;
    if (jj_scan_token(47)) {
    jj_scanpos = xsp;
    if (jj_scan_token(48)) {
    jj_scanpos = xsp;
    if (jj_scan_token(50)) {
    jj_scanpos = xsp;
    if (jj_scan_token(51)) {
    jj_scanpos = xsp;
    if (jj_scan_token(52)) {
    jj_scanpos = xsp;
    if (jj_scan_token(53)) {
    jj_scanpos = xsp;
    if (jj_scan_token(54)) {
    jj_scanpos = xsp;
    if (jj_scan_token(55)) {
    jj_scanpos = xsp;
    if (jj_scan_token(56)) {
    jj_scanpos = xsp;
    if (jj_scan_token(57)) {
    jj_scanpos = xsp;
    if (jj_scan_token(58)) {
    jj_scanpos = xsp;
    if (jj_scan_token(59)) {
    jj_scanpos = xsp;
    if (jj_scan_token(60)) {
    jj_scanpos = xsp;
    if (jj_scan_token(61)) {
    jj_scanpos = xsp;
    if (jj_scan_token(62)) {
    jj_scanpos = xsp;
    if (jj_scan_token(63)) {
    jj_scanpos = xsp;
    if (jj_scan_token(64)) {
    jj_scanpos = xsp;
    if (jj_scan_token(65)) {
    jj_scanpos = xsp;
    if (jj_scan_token(66)) {
    jj_scanpos = xsp;
    if (jj_scan_token(67)) {
    jj_scanpos = xsp;
    if (jj_scan_token(68)) {
    jj_scanpos = xsp;
    if (jj_scan_token(69)) {
    jj_scanpos = xsp;
    if (jj_scan_token(70)) {
    jj_scanpos = xsp;
    if (jj_scan_token(71)) {
    jj_scanpos = xsp;
    if (jj_scan_token(31)) return true;
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    }
    return false;
  }

  private boolean jj_3R_12() {
    if (jj_3R_22()) return true;
    return false;
  }

  private boolean jj_3R_33() {
    if (jj_3R_38()) return true;
    return false;
  }

  private boolean jj_3R_21() {
    if (jj_scan_token(BANG)) return true;
    return false;
  }

  private boolean jj_3R_27() {
    if (jj_3R_18()) return true;
    return false;
  }

  private boolean jj_3R_11() {
    if (jj_3R_21()) return true;
    return false;
  }

  private boolean jj_3R_41() {
    if (jj_scan_token(COMMA)) return true;
    if (jj_3R_15()) return true;
    return false;
  }

  private boolean jj_3R_20() {
    if (jj_scan_token(VAR_IDENTIFIER)) return true;
    return false;
  }

  private boolean jj_3R_17() {
    Token xsp;
    xsp = jj_scanpos;
    if (jj_scan_token(23)) {
    jj_scanpos = xsp;
    if (jj_scan_token(21)) {
    jj_scanpos = xsp;
    if (jj_scan_token(37)) {
    jj_scanpos = xsp;
    if (jj_scan_token(31)) {
    jj_scanpos = xsp;
    if (jj_scan_token(72)) return true;
    }
    }
    }
    }
    xsp = jj_scanpos;
    if (jj_3R_33()) jj_scanpos = xsp;
    return false;
  }

  private boolean jj_3R_26() {
    if (jj_3R_17()) return true;
    return false;
  }

  private boolean jj_3R_37() {
    if (jj_3R_20()) return true;
    return false;
  }

  private boolean jj_3R_44() {
    if (jj_3R_20()) return true;
    return false;
  }

  private boolean jj_3R_10() {
    if (jj_3R_20()) return true;
    return false;
  }

  private boolean jj_3R_43() {
    if (jj_3R_23()) return true;
    return false;
  }

  private boolean jj_3R_19() {
    Token xsp;
    xsp = jj_scanpos;
    if (jj_scan_token(15)) {
    jj_scanpos = xsp;
    if (jj_scan_token(19)) return true;
    }
    return false;
  }

  private boolean jj_3R_9() {
    if (jj_3R_19()) return true;
    return false;
  }

  private boolean jj_3R_36() {
    if (jj_3R_22()) return true;
    return false;
  }

  private boolean jj_3R_42() {
    if (jj_scan_token(BAR)) return true;
    Token xsp;
    xsp = jj_scanpos;
    if (jj_3R_43()) {
    jj_scanpos = xsp;
    if (jj_3R_44()) return true;
    }
    return false;
  }

  private boolean jj_3R_39() {
    if (jj_3R_15()) return true;
    Token xsp;
    while (true) {
      xsp = jj_scanpos;
      if (jj_3R_41()) { jj_scanpos = xsp; break; }
    }
    xsp = jj_scanpos;
    if (jj_3R_42()) jj_scanpos = xsp;
    return false;
  }

  private boolean jj_3R_25() {
    if (jj_3R_16()) return true;
    return false;
  }

  private boolean jj_3R_8() {
    if (jj_3R_18()) return true;
    return false;
  }

  private boolean jj_3R_32() {
    if (jj_3R_23()) return true;
    return false;
  }

  private boolean jj_3R_35() {
    if (jj_3R_17()) return true;
    return false;
  }

  private boolean jj_3R_23() {
    if (jj_scan_token(LBRACKET)) return true;
    Token xsp;
    xsp = jj_scanpos;
    if (jj_3R_39()) jj_scanpos = xsp;
    if (jj_scan_token(RBRACKET)) return true;
    return false;
  }

  private boolean jj_3R_7() {
    if (jj_3R_17()) return true;
    return false;
  }

  private boolean jj_3R_40() {
    if (jj_scan_token(COMMA)) return true;
    if (jj_3R_15()) return true;
    return false;
  }

  private boolean jj_3_2() {
    if (jj_3R_5()) return true;
    return false;
  }

  private boolean jj_3R_31() {
    if (jj_3R_22()) return true;
    return false;
  }

  /** Generated Token Manager. */
  public IntPrologTokenManager token_source;
  SimpleCharStream jj_input_stream;
  /** Current token. */
  public Token token;
  /** Next token. */
  public Token jj_nt;
  private int jj_ntk;
  private Token jj_scanpos, jj_lastpos;
  private int jj_la;
  private int jj_gen;
  final private int[] jj_la1 = new int[16];
  static private int[] jj_la1_0;
  static private int[] jj_la1_1;
  static private int[] jj_la1_2;
  static {
      jj_la1_init_0();
      jj_la1_init_1();
      jj_la1_init_2();
   }
   private static void jj_la1_init_0() {
      jj_la1_0 = new int[] {0x80a00000,0x80a00000,0x0,0x0,0x80a00000,0x2000000,0x83a00000,0x0,0xa3a88000,0x0,0x21000000,0x0,0xa3a88000,0x88000,0x80000000,0xa3a88000,};
   }
   private static void jj_la1_init_1() {
      jj_la1_1 = new int[] {0x22,0x20,0x1,0x0,0x20,0x0,0x20,0x1,0x20034,0x1,0x0,0x0,0x20034,0x0,0xfffdfff8,0x20034,};
   }
   private static void jj_la1_init_2() {
      jj_la1_2 = new int[] {0x180,0x100,0x0,0x80,0x100,0x0,0x100,0x0,0x100,0x0,0x0,0x100,0x100,0x0,0xff,0x100,};
   }
  final private JJCalls[] jj_2_rtns = new JJCalls[2];
  private boolean jj_rescan = false;
  private int jj_gc = 0;

  /** Constructor with InputStream. */
  public IntProlog(java.io.InputStream stream) {
     this(stream, null);
  }
  /** Constructor with InputStream and supplied encoding */
  public IntProlog(java.io.InputStream stream, String encoding) {
    try { jj_input_stream = new SimpleCharStream(stream, encoding, 1, 1); } catch(java.io.UnsupportedEncodingException e) { throw new RuntimeException(e); }
    token_source = new IntPrologTokenManager(jj_input_stream);
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 16; i++) jj_la1[i] = -1;
    for (int i = 0; i < jj_2_rtns.length; i++) jj_2_rtns[i] = new JJCalls();
  }

  /** Reinitialise. */
  public void ReInit(java.io.InputStream stream) {
     ReInit(stream, null);
  }
  /** Reinitialise. */
  public void ReInit(java.io.InputStream stream, String encoding) {
    try { jj_input_stream.ReInit(stream, encoding, 1, 1); } catch(java.io.UnsupportedEncodingException e) { throw new RuntimeException(e); }
    token_source.ReInit(jj_input_stream);
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 16; i++) jj_la1[i] = -1;
    for (int i = 0; i < jj_2_rtns.length; i++) jj_2_rtns[i] = new JJCalls();
  }

  /** Constructor. */
  public IntProlog(java.io.Reader stream) {
    jj_input_stream = new SimpleCharStream(stream, 1, 1);
    token_source = new IntPrologTokenManager(jj_input_stream);
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 16; i++) jj_la1[i] = -1;
    for (int i = 0; i < jj_2_rtns.length; i++) jj_2_rtns[i] = new JJCalls();
  }

  /** Reinitialise. */
  public void ReInit(java.io.Reader stream) {
    jj_input_stream.ReInit(stream, 1, 1);
    token_source.ReInit(jj_input_stream);
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 16; i++) jj_la1[i] = -1;
    for (int i = 0; i < jj_2_rtns.length; i++) jj_2_rtns[i] = new JJCalls();
  }

  /** Constructor with generated Token Manager. */
  public IntProlog(IntPrologTokenManager tm) {
    token_source = tm;
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 16; i++) jj_la1[i] = -1;
    for (int i = 0; i < jj_2_rtns.length; i++) jj_2_rtns[i] = new JJCalls();
  }

  /** Reinitialise. */
  public void ReInit(IntPrologTokenManager tm) {
    token_source = tm;
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 16; i++) jj_la1[i] = -1;
    for (int i = 0; i < jj_2_rtns.length; i++) jj_2_rtns[i] = new JJCalls();
  }

  private Token jj_consume_token(int kind) throws ParseException {
    Token oldToken;
    if ((oldToken = token).next != null) token = token.next;
    else token = token.next = token_source.getNextToken();
    jj_ntk = -1;
    if (token.kind == kind) {
      jj_gen++;
      if (++jj_gc > 100) {
        jj_gc = 0;
        for (int i = 0; i < jj_2_rtns.length; i++) {
          JJCalls c = jj_2_rtns[i];
          while (c != null) {
            if (c.gen < jj_gen) c.first = null;
            c = c.next;
          }
        }
      }
      return token;
    }
    token = oldToken;
    jj_kind = kind;
    throw generateParseException();
  }

  static private final class LookaheadSuccess extends java.lang.Error { }
  final private LookaheadSuccess jj_ls = new LookaheadSuccess();
  private boolean jj_scan_token(int kind) {
    if (jj_scanpos == jj_lastpos) {
      jj_la--;
      if (jj_scanpos.next == null) {
        jj_lastpos = jj_scanpos = jj_scanpos.next = token_source.getNextToken();
      } else {
        jj_lastpos = jj_scanpos = jj_scanpos.next;
      }
    } else {
      jj_scanpos = jj_scanpos.next;
    }
    if (jj_rescan) {
      int i = 0; Token tok = token;
      while (tok != null && tok != jj_scanpos) { i++; tok = tok.next; }
      if (tok != null) jj_add_error_token(kind, i);
    }
    if (jj_scanpos.kind != kind) return true;
    if (jj_la == 0 && jj_scanpos == jj_lastpos) throw jj_ls;
    return false;
  }


/** Get the next Token. */
  final public Token getNextToken() {
    if (token.next != null) token = token.next;
    else token = token.next = token_source.getNextToken();
    jj_ntk = -1;
    jj_gen++;
    return token;
  }

/** Get the specific Token. */
  final public Token getToken(int index) {
    Token t = token;
    for (int i = 0; i < index; i++) {
      if (t.next != null) t = t.next;
      else t = t.next = token_source.getNextToken();
    }
    return t;
  }

  private int jj_ntk() {
    if ((jj_nt=token.next) == null)
      return (jj_ntk = (token.next=token_source.getNextToken()).kind);
    else
      return (jj_ntk = jj_nt.kind);
  }

  private java.util.List<int[]> jj_expentries = new java.util.ArrayList<int[]>();
  private int[] jj_expentry;
  private int jj_kind = -1;
  private int[] jj_lasttokens = new int[100];
  private int jj_endpos;

  private void jj_add_error_token(int kind, int pos) {
    if (pos >= 100) return;
    if (pos == jj_endpos + 1) {
      jj_lasttokens[jj_endpos++] = kind;
    } else if (jj_endpos != 0) {
      jj_expentry = new int[jj_endpos];
      for (int i = 0; i < jj_endpos; i++) {
        jj_expentry[i] = jj_lasttokens[i];
      }
      jj_entries_loop: for (java.util.Iterator<?> it = jj_expentries.iterator(); it.hasNext();) {
        int[] oldentry = (int[])(it.next());
        if (oldentry.length == jj_expentry.length) {
          for (int i = 0; i < jj_expentry.length; i++) {
            if (oldentry[i] != jj_expentry[i]) {
              continue jj_entries_loop;
            }
          }
          jj_expentries.add(jj_expentry);
          break jj_entries_loop;
        }
      }
      if (pos != 0) jj_lasttokens[(jj_endpos = pos) - 1] = kind;
    }
  }

  /** Generate ParseException. */
  public ParseException generateParseException() {
    jj_expentries.clear();
    boolean[] la1tokens = new boolean[73];
    if (jj_kind >= 0) {
      la1tokens[jj_kind] = true;
      jj_kind = -1;
    }
    for (int i = 0; i < 16; i++) {
      if (jj_la1[i] == jj_gen) {
        for (int j = 0; j < 32; j++) {
          if ((jj_la1_0[i] & (1<<j)) != 0) {
            la1tokens[j] = true;
          }
          if ((jj_la1_1[i] & (1<<j)) != 0) {
            la1tokens[32+j] = true;
          }
          if ((jj_la1_2[i] & (1<<j)) != 0) {
            la1tokens[64+j] = true;
          }
        }
      }
    }
    for (int i = 0; i < 73; i++) {
      if (la1tokens[i]) {
        jj_expentry = new int[1];
        jj_expentry[0] = i;
        jj_expentries.add(jj_expentry);
      }
    }
    jj_endpos = 0;
    jj_rescan_token();
    jj_add_error_token(0, 0);
    int[][] exptokseq = new int[jj_expentries.size()][];
    for (int i = 0; i < jj_expentries.size(); i++) {
      exptokseq[i] = jj_expentries.get(i);
    }
    return new ParseException(token, exptokseq, tokenImage);
  }

  /** Enable tracing. */
  final public void enable_tracing() {
  }

  /** Disable tracing. */
  final public void disable_tracing() {
  }

  private void jj_rescan_token() {
    jj_rescan = true;
    for (int i = 0; i < 2; i++) {
    try {
      JJCalls p = jj_2_rtns[i];
      do {
        if (p.gen > jj_gen) {
          jj_la = p.arg; jj_lastpos = jj_scanpos = p.first;
          switch (i) {
            case 0: jj_3_1(); break;
            case 1: jj_3_2(); break;
          }
        }
        p = p.next;
      } while (p != null);
      } catch(LookaheadSuccess ls) { }
    }
    jj_rescan = false;
  }

  private void jj_save(int index, int xla) {
    JJCalls p = jj_2_rtns[index];
    while (p.gen > jj_gen) {
      if (p.next == null) { p = p.next = new JJCalls(); break; }
      p = p.next;
    }
    p.gen = jj_gen + xla - jj_la; p.first = token; p.arg = xla;
  }

  static final class JJCalls {
    int gen;
    Token first;
    int arg;
    JJCalls next;
  }

}
