package aplprolog.prolog; 

import java.io.FileInputStream;
import java.io.StringReader;  
import java.util.ArrayList;

import aplprolog.Parsing.ParseException;
import aplprolog.datastructures.PrimIntArrayList;
import aplprolog.datastructures.PrimIntKeyHashMap;  
/**
 * Main class of IntProlog, contains basic resolution steps and overhead variables.
 * 
 * @author Bas Testerink, Utrecht University, The Netherlands
 *
 */
public class Prolog {
	public boolean trace = false, traceSol = false;
	public static int MOVETOPS = 0,SHAREVAR = 1, BINDVAR = 2, MAKEVAR = 3, ADDGOALS = 4, MOVERULECURSOR = 5, ADDQUERY = 6, CUTMARKER = 7, COLLECTORID=8, TOBEDELETED=Integer.MIN_VALUE; 
	public static int[] intsize = new int[]{1,3,2,2,2,2,1,2,2}; 
	public PrimIntArrayList history = new PrimIntArrayList(true);
	public PrimIntArrayList todo = new PrimIntArrayList(false), processed = new PrimIntArrayList(false);
	public PrimIntKeyHashMap<PrimIntKeyHashMap<PrimIntArrayList>> rulebase = new PrimIntKeyHashMap<PrimIntKeyHashMap<PrimIntArrayList>>(PrimIntKeyHashMap.OBJ);
	public StringStorage strStorage = new StringStorage(); 
	public IntHarvester harvester;
	public Substitution substitution;
	public aplprolog.Parsing.IntProlog mp; 
	public BuiltIn builtin; 
	int cutmark = 0;
	public boolean rulesModified = false; // keeps track of whether the last query changed the rulebase (needed for 2OPL)
	public String solStr = ""; // for when tracing solutions
	ArrayList<String> shadowPredicates = new ArrayList<String>(); // for output, library predicates are not shown in the toString() string
	public int lastcursor=0; // last tried rulecursor, not nice but is a quick workaround

	public Prolog(){  
		harvester = new IntHarvester(this);
		substitution = new Substitution(this);
		builtin = new BuiltIn(this); 
		mp = new aplprolog.Parsing.IntProlog(new StringReader(""));
	}

	public static void main(String[] args){ 

		Prolog p = new Prolog();
		p.loadfile("experiment.pl", true);
		p.traceSol = true;
		//p.trace = true;
		//p.query("listcount([1,2,3,4,5,6,7,8,9,0,1,2,3,4,5,6,7,8,9,0,2,3,4,5,6,7,8,9,0,1,2,3,4,5,6,7,8,9,0],R).");
		long l = System.currentTimeMillis();
		for(int i = 0; i < 1; i++)p.query("einstein(A,B).");
		System.out.println(System.currentTimeMillis()-l);
		System.out.println(p.solStr);
		//System.out.println(p.processed.pointer);
		try {   
			//p.trace = true;
			//p.loadfile("harry.pl");
			//p.loadfile("piet.pl");
			//p.traceSol = true;
			//String query = "";
			//p.loadfile("experiment.pl");  
			// query = "listcount([1,2,3,4,5,6,7,8,9,0,1,2,3,4,5,6,7,8,9,0],R).";  
			// query = "einstein(Houses, FishOwner).";
			// query = "findall(Y,(pakY(Y,X),findall(p(X),Y,[p(dit),p(is),p(een),p(test)|[]])),L),gelijk([a,b,c],[C|[D|E]],[A|B]),einstein(Houses, FishOwner).";
			// query = "f(X),h(X)."; 

			//p.testquery(query);
			//if(p.trace)System.out.println("%query: "+query);
			//System.out.println(p.query(query));
			//p.mp.queryVars.clear();
			//p.testcase();
			//p.modificationTest();
		} catch (Exception e) {
			e.printStackTrace();
		} 
	} 

	/**
	 * Perform string query.
	 * @param q Query (string format).
	 * @return Whether the query was provable.
	 */
	public boolean query(String q){
		int[] query = null;
		solStr = "";
		try{
			query = mp.parseQuery(q, strStorage);
		} catch(ParseException pe){ 
			if(traceSol){
				String addition = "true :- ";
				int additionLength = addition.length();
				solStr = "Parse error. ";
				solStr += "Last successful token: \""+pe.currentToken.image+"\", ending at position "+(pe.currentToken.endColumn-additionLength)+".\r\n";
				solStr += (q.substring(0,pe.currentToken.endColumn-additionLength)+"**ERROR**"+q.substring(pe.currentToken.endColumn-additionLength))+"\r\n";
			}
			return false;
		}
		boolean r = query(query);
		if(!r&&traceSol) solStr += "false (no solutions found)";
		return r;
	}
	/**
	 * Perform integer array query.
	 * @param query Query (integer format).
	 * @return Whether the query was provable.
	 */
	public synchronized boolean query(int[] query){  
		//trace=true;
		history.clear();
		todo.clear();
		substitution.clear(); 
		processed.clear();  
		harvester.buildLocalVars(query);
		rulesModified = false;
		substitution.id = query[query.length-query[query.length-1]-2]+2;
		PrimIntArrayList subgoals = harvester.getBody2(query);
		history.addInt(subgoals.pointer);
		history.addInt(ADDGOALS);
		while(!subgoals.isEmpty())todo.addAr(subgoals.removeLastAr());
		history.addInt(ADDQUERY);  
		substitution.id = 10000+2; 
		boolean r = prove();
		removeToBeDeleted();
		return r;
	}

	/**
	 * First time query for a solution iterator. Requires less cleanup beforehand.
	 * @param query Query (integer format).
	 * @return Whether the query is provable.
	 */
	public boolean firstTimeQuery(int[] query){  
		harvester.buildLocalVars(query);
		substitution.id = query[query.length-query[query.length-1]-2]+2;
		PrimIntArrayList subgoals = harvester.getBody2(query);
		history.addInt(subgoals.pointer);
		history.addInt(ADDGOALS);
		while(!subgoals.isEmpty())todo.addAr(subgoals.removeLastAr());
		history.addInt(ADDQUERY); 
		return prove(); 
	}

	/**
	 * Load a rule base from a prolog file.
	 * @param file Path to file that contains the rules.
	 */
	public void loadfile(String file, boolean shadow){
		try{
			mp.ReInit(new FileInputStream(file));
			mp.parseRulebase(strStorage);
			takeOverRules(mp.rules, shadow);
		} catch(Exception e){
			e.printStackTrace();
		}
	}
	/**
	 * Take over the rulebase from a parsing result. Over writes any existing rules.
	 * @param rules The rule base to be adapted.
	 */
	public void takeOverRules(PrimIntKeyHashMap<PrimIntKeyHashMap<PrimIntArrayList>> rules, boolean shadow){
		if(rules.size() > 0){
			for(int i1 = rules.firstKey(); rules.hasNext(); i1 = rules.nextKey()){
				for(int i2 = rules.getObj(i1).firstKey(); rules.getObj(i1).hasNext(); i2 = rules.getObj(i1).nextKey()){
					if(trace)System.out.println(strStorage.getString(i1)+" "+i1+" "+i2+" "+shadow);
					if(rulebase.getObj(i1)==null)
						rulebase.putObj(i1,new PrimIntKeyHashMap<PrimIntArrayList>(PrimIntKeyHashMap.OBJ));
					if(rulebase.getObj(i1).getObj(i2)==null)
						rulebase.getObj(i1).putObj(i2, new PrimIntArrayList(false));
					PrimIntArrayList ars = rules.getObj(i1).getObj(i2);
					if(shadow)shadowPredicates.add(strStorage.getString(i1)+"/"+i2); // add to shadow base
					for(int i = 0; i < ars.pointer; i++)
						rulebase.getObj(i1).getObj(i2).addAr(ars.arraydata[i]);
				}
			}
		}
	}
	/**
	 * Basic resolution call. Tries to prove the current set of to-do subgoals.
	 * @return Whether all goals are provable.
	 */
	public boolean prove(){
		while(!todo.isEmpty()){
			if(trace)System.out.println("-----new prove cycle, top todo: "+arStr(todo.getLastAr())+" interpreted: "+harvester.arToString(todo.getLastAr(), false));
			if(history.getLastInt()!=MOVERULECURSOR){ // if you have not just backtracked then the last history int must be "MOVETOPS"
				history.addInt(-1); // reset rule cursor 
				history.addInt(MOVERULECURSOR); // so add decision point
			}
			if(!findNextRule(todo.getLastAr(),false)){
				if(trace)System.out.println(": backtrack because there is no rule (left) for the current goal");
				backtrack(true); // otherwise backtrack
				if(history.getLastInt()==ADDQUERY) return false;
			} else{
				if(!todo.isEmpty()) // TODO: without it we get error with 2OPL, check out why
					fromToDoToProcessed();
			}
		} 
		if(trace||traceSol){
			for(String j : mp.vars.keySet()){
				//System.out.println(j+" = "+harvester.arToString(substitution.value(mp.vars.get(j)+1),false));
				solStr += j+" = "+harvester.arToString(substitution.value(mp.vars.get(j)+1),false) +"\r\n";
			}
		}
		return true;
	}
	/**
	 * Prove a subgoal.
	 * @param initcall The subgoal.
	 * @return Whether there are still applicable rules left.
	 */
	public boolean findNextRule(int[] initcall, boolean callInVar){
		int[] call = initcall; 
		if(call[0]==IntHarvester.CUT){
			history.removeLastInt(); // remove the fact that this is a decision point, because it isn't
			history.removeLastInt(); 
			history.addInt(call[1]);
			history.addInt(CUTMARKER); // put the marker in place, the next time it gets backtrack it will undo everything up till the sibling marker
			if(trace)System.out.println(": adding second cut-mark with id "+call[1]);
			return true; // continue
		} else if(call[0]==IntHarvester.VARIABLE){
			int[] value = substitution.value(call[1]);
			if(value == null) return false;
			return findNextRule(value,true);
		} else if(call[0]==IntHarvester.PARENTHESIZED){
			if(history.intdata[history.pointer-2]==-1){ // can't backtrack on putting parenthesized goals in the stack
				history.intdata[history.pointer-2] = 0;
				addSubGoals(harvester.getBody3(call, 2));
				return true;
			} else return false;
		} else if(call[0] != IntHarvester.PREDICATE) return false;
		// update the cursor
		int test = history.pointer;
		int cursor = ++history.intdata[test-2]; 
		// catch built in predicates
		if(call[1]<builtin.range)
			if(call[2]==BuiltIn.arity[call[1]])
				return builtin.execute(call,callInVar);
		// get the candidate rule
		int[] rule = ruleCursorCheck(cursor,call);// get the next candidate
		history.intdata[test-2] = lastcursor;
		if(rule!=null){
			if(trace)System.out.println(": trying rule: "+harvester.ruleToString(rule)); 
			if(!headmatch(call,rule,callInVar)){ // if the head does not match the call
				if(trace)System.out.println(": backtrack on call/head unification");
				backtrack(false); // then undo all variable changes etc, because skip is false, it will stop exactly at the state at which the process was when the call was initiated
				return findNextRule(call,false); // retry (will move cursor forwards)
			} else {
				addSubGoals(rule); // copy the goals onto the todo stack BEFORE the current todo top (this one)
				return true;
			}
		}
		return false; // there is no rule to apply
	}
	/**
	 * Backtrack call.
	 * @param skip Whether to keep going backwards at choice points.
	 */
	public void backtrack(boolean skip){
		while(undo(skip,false)); // go back to decision point
	}
	/**
	 * Cut backtrack blindly backtracks untill the appropriate cutmarker is hit.
	 * @param marker Id of the marker to look for.
	 * @param skip Whether to skip at choice points.
	 */
	public void cutbacktrack(int marker,boolean skip){
		if(trace)System.out.println(": undoing everything between the two cut-marks with id "+(marker-1)+" and "+marker);
		while(history.getLastInt()!=CUTMARKER||history.intdata[history.pointer-2]!=marker-1)
			undo(skip,true); // undo everything between the two markers
		history.removeLastInt(); // remove the first marker
		history.removeLastInt();
		int c = history.pointer-1; // you don't want to backtrack on the parent if you are backtracking on a cut
		while(history.intdata[c]!=MOVERULECURSOR) c -= intsize[history.intdata[c]];
		history.removeIntAt(c-1);
		history.addInt(c-1,Integer.MAX_VALUE); // make sure the parent isn't reconsidered
		if(trace)System.out.println(": everything between the two cut-marks with id "+(marker-1)+" and "+marker+" is undone");
	}
	/**
	 * Check to see whether there is a candidate rule left for a goal.
	 * @param cursor Cursor in the call.
	 * @param call The call.
	 * @return Whether there is a rule left.
	 */
	public int[] ruleCursorCheck(int cursor, int[] call){		

		lastcursor = cursor;
		if(call[0]==IntHarvester.PREDICATE){
			if(call[1] == BuiltIn.IC_FINDALL){
				if(call[2] == 3){
					return cursor==1?null:call;
				}
			}
		}

		if(call[1] == builtin.IC_RESETSTABLE)
			if(call[2]==BuiltIn.arity[call[1]])
				return call; //TODO is a hack workaround for built ins which you can backtrack on

		if(call[0]==IntHarvester.VARIABLE){
			int[] value = substitution.value(call[1]);
			if(value == null) return null;
			if(value[0]==IntHarvester.PREDICATE)
				call = value;
			else return null;
		}
		int[] r = null;
		PrimIntKeyHashMap<PrimIntArrayList> idEntry = rulebase.getObj(call[1]);
		if(idEntry!=null){ // there are entries for the id
			PrimIntArrayList rules = idEntry.getObj(call[2]); 
			if(rules != null) { // there are entries for the (id,arity)-combo
				while((r==null||r[0]==TOBEDELETED)&&cursor<rules.pointer){ // you haven't tried all rules yet
					r = rules.getAr(cursor);
					lastcursor = cursor;
					cursor++;
				}
			}
		}
		if(r!=null&&r[0]==TOBEDELETED) r = null;
		if(trace){
			String out = ": no rules available for "+harvester.arToString(call, false)+" ";
			if(rulebase.getObj(call[1])==null){ out+="(no entry for the predicate id)";} // no entry for the id
			else if(rulebase.getObj(call[1]).getObj(call[2]) == null){ out+="(no entry for the predicate arity)";} // no entry or the arg count
			else if(cursor >= rulebase.getObj(call[1]).getObj(call[2]).pointer){ out+=(cursor==Integer.MAX_VALUE?"(due to a cut)":"(all rules have been tried)");}
			if(r==null)System.out.println(out);
		}
		return r;
	}
	/**
	 * Move a to-do goal to the 'processed' stack.
	 */
	public void fromToDoToProcessed(){ 
		processed.addAr(todo.removeLastAr()); // move the top from the todo stack to the processed stack
		history.addInt(MOVETOPS);
		if(trace)System.out.println(": moving "+harvester.arToString(processed.getLastAr(), false)+" from todo to processed");
	}
	/**
	 * Undo a resolution action.
	 * @param skip Whether to call backtrack at a choice point.
	 * @param inCut Whether the algorithm is inside a cut backtrack.
	 * @return Whether the last action can be undone.
	 */
	public boolean undo(boolean skip, boolean inCut){ // skip stands for: when you encounter a decision point which has no more rules available, will you skip it?
		if(history.pointer == 0) return false;
		int action = history.removeLastInt();
		if(action == MOVETOPS){
			todo.addAr(processed.removeLastAr());
			if(trace)System.out.println(": moving "+harvester.arToString(todo.getLastAr(), false)+" from processed to todo");
		} else if(action == BINDVAR){
			substitution.free(history.removeLastInt());
		} else if(action == SHAREVAR){
			substitution.unshare(history.removeLastInt(), history.removeLastInt());
		} else if(action == MAKEVAR){
			substitution.delete(history.removeLastInt());
		} else if(action == ADDGOALS){
			int j = history.removeLastInt();
			if(trace)System.out.print(": removing from todo list: ");
			for(int i = 0; i < j; i++){
				if(trace)System.out.print(harvester.arToString(todo.removeArAt(todo.pointer-2), false)); // "-2" because you add goals and then put the top todo on the processed stack, so when backtracking the supergoal is put on top of his subgoals
				else todo.removeArAt(todo.pointer-2);
			}
			if(trace)System.out.println();
		} else if(action == MOVERULECURSOR){
			if(inCut) history.removeLastInt(); // don't need to re-check decision points
			else if(skip){
				int cursor = history.removeLastInt(); 
				if(ruleCursorCheck(cursor,todo.getLastAr())==null) return true; // if that decision point has no siblings; backtracking should continue
				history.addInt(cursor); // still rules possible
				history.addInt(MOVERULECURSOR);
			} else history.addInt(MOVERULECURSOR);
			return false; // stop backtracking
		} else if(action == ADDQUERY){
			history.addInt(ADDQUERY);
			return false;
		} else if(action == CUTMARKER){
			int mark = history.removeLastInt();
			if(mark%2 == 1&&!inCut)	cutbacktrack(mark, skip);
		} else if(action == COLLECTORID){
			builtin.answersets.collections.removeObj(history.removeLastInt());
		}
		return true;
	}
	/**
	 * Unify the head of a rule with a subgoal.
	 * @param call The subgoal.
	 * @param rule The rule to unify with.
	 * @param callInVar Whether the subgoal is inside a variable.
	 * @return Whether the head of the rule was unifiable with the subgoal.
	 */
	public boolean headmatch(int[] call, int[] rule, boolean callInVar){ // does the call match the head of the rule?
		harvester.buildLocalVars(rule);
		return harvester.unifyTerm(call, rule, 0, 0,callInVar,false);
	}

	/**
	 * Add the subgoals of a rule to the todo stack.
	 * @param rule Rule to be added.
	 */
	public void addSubGoals(int[] rule){
		PrimIntArrayList subgoals = harvester.getBody2(rule);
		addSubGoals(subgoals);
	}
	/**
	 * Add a list of subgoals to the todo stack.
	 * @param subgoals Goals to be added.
	 */
	public void addSubGoals(PrimIntArrayList subgoals){
		if(subgoals.pointer>0){
			if(trace){
				System.out.print(": adding "+subgoals.pointer+" goal(s). That/Those being: ");

				for(int i = 0; i < subgoals.pointer; i++)
					System.out.print(harvester.arToString(subgoals.getAr(i), false)+" ");

				System.out.println();
			}
			int[] current = todo.removeLastAr();
			boolean markmade = false;
			int mark = 0;
			history.addInt(subgoals.pointer);
			history.addInt(ADDGOALS);
			while(!subgoals.isEmpty()){
				int[] sub = subgoals.removeLastAr();
				if(sub[0] == IntHarvester.CUT){
					if(!markmade){
						markmade = true;
						mark = nextCutMark();
						history.addInt(mark);
						history.addInt(CUTMARKER);
						if(trace)System.out.println(": adding first cut-mark with id " + mark);
					}
					sub[1] = mark+1;
				}
				todo.addAr(sub);
			}
			todo.addAr(current);
		}
	}

	/**
	 * Generate the next cut mark id.
	 * @return The next cut mark.
	 */
	public int nextCutMark(){
		cutmark = (cutmark+2)%Integer.MAX_VALUE;
		return cutmark;
	}

	/**
	 * Removes all the rules that are to be deleted.
	 */
	public void removeToBeDeleted(){
		for(int i1 = rulebase.firstKey(); rulebase.hasNext(); i1 = rulebase.nextKey()){
			for(int i2 = rulebase.getObj(i1).firstKey(); rulebase.getObj(i1).hasNext(); i2 = rulebase.getObj(i1).nextKey()){
				PrimIntArrayList list = rulebase.getObj(i1).getObj(i2);
				for(int i = list.pointer-1; i >= 0; i--)
					if(list.arraydata[i][0]==TOBEDELETED)
						list.removeArAt(i);
			}
		}
	}

	/**
	 * Convert an integer list to a primitive integer array.
	 * @param array List to be transformed.
	 * @return Primitive representation of the list.
	 */
	public static int[] listToPrim(PrimIntArrayList array){
		int[] prim = new int[array.pointer]; 
		for(int i = 0; i < prim.length; i++) prim[i] = array.intdata[i]; // make primitive version
		return prim;
	}
	/**
	 * Extract an array from another array.
	 * @param array Extraction array.
	 * @param a Start point.
	 * @param b End point.
	 * @return The array between a and b.
	 */
	public int[] extractArray(int[] array, int a, int b){
		int[] r = new int[b-a];
		for(int i = 0; i < r.length; i++){
			r[i] = array[a+i];
		}
		return r;
	}
	/**
	 * String representation of a primitive array.
	 * @param ar Array to convert.
	 * @return String representation.
	 */
	public String arStr(int[] ar){
		if(ar==null) return "NULL";
		String s = "[";
		for(int i = 0; i < ar.length; i++){
			s+=ar[i];
			if(i!=ar.length-1)s+=",";
		}
		return s + "]";
	}

	public String toString(){
		String r = "";
		for(int i1 = rulebase.firstKey(); rulebase.hasNext(); i1 = rulebase.nextKey()){
			String s = "";
			boolean show = false;
			PrimIntKeyHashMap<PrimIntArrayList> name = rulebase.getObj(i1);
			if(name.size()>0){ // do not show empty name entries
				s += "% Rules with string id: "+strStorage.getString(i1)+"\r\n";
				for(int i2 = name.firstKey(); name.hasNext(); i2 = name.nextKey()){
					PrimIntArrayList arity = name.getObj(i2);
					String name_arity = strStorage.getString(i1)+"/"+i2;
					if(arity.pointer>0 && !shadowPredicates.contains(name_arity)){ // do not show shadowed or empty pred/arity pairs
						s += "% "+name_arity+":\r\n";
						for(int i3 = 0; i3 < arity.pointer; i3++){
							s += harvester.ruleToString(arity.arraydata[i3])+"\r\n";
							show = true;
						}
					}
				}
				s += "\r\n";
			}
			if(show)r+=s;
		}
		return r;
	}
}
