package aplprolog.prolog;

import java.util.HashMap;
import aplprolog.datastructures.PrimIntKeyHashMap;
/**
 * This class implements a mapping from Strings to integers and back. Basically it is just two hash maps.
 * 
 * @author Bas Testerink, Utrecht University, The Netherlands
 * 
 */
public class StringStorage {
	private PrimIntKeyHashMap<String> intToString = new PrimIntKeyHashMap<String>(PrimIntKeyHashMap.OBJ);
	private HashMap<String,Integer> stringToInt = new HashMap<String, Integer>();
	private int counter = 0;
	
	public StringStorage(){
	}
	
	/**
	 * Get a string.
	 * @param id Id of the string to be returned.
	 * @return String that belongs to the id.
	 */
	public String getString(int id){
		return intToString.getObj(id);
	}
	/**
	 * Get an int.
	 * @param id Id of the int to be returned.
	 * @return Int that belongs to the id.
	 */
	public Integer getInt(String id){
		return stringToInt.get(id);
	}
	
	/**
	 * Remove a string/int pair.
	 * @param id Id of the string to be removed. 
	 */
	public void remove(int id){
		String str = intToString.getObj(id);
		intToString.removeObj(id);
		stringToInt.remove(str);
	}
	/**
	 * Remove a string/int pair.
	 * @param id Id of the int to be removed. 
	 */
	public void remove(String id){
		Integer integer = stringToInt.get(id);
		stringToInt.remove(id);
		intToString.removeObj(integer);
	}
	/**
	 * Add a string.
	 * @param s String to be added.
	 * @return Corresponding int id.
	 */
	public int add(String s){
		Integer id = counter;
		counter++;
		intToString.putObj(id, s);
		stringToInt.put(s, id);
		return id;
	}
	
	/**
	 * Replace a string.
	 * @param oldStr Old string.
	 * @param newStr Replacement.
	 */
	public void replaceString(String oldStr, String newStr){
		Integer id = stringToInt.get(oldStr);
		stringToInt.put(newStr, id);
		intToString.removeObj(id);
		intToString.putObj(id, newStr);
	}
}
