package aplprolog.prolog.builtins;
import java.util.*;

import aplprolog.datastructures.PrimIntArrayList;
import aplprolog.datastructures.PrimIntKeyHashMap;
import aplprolog.prolog.*;
/**
 * Built in functionalities concerning answer sets.
 * 
 * @author Bas Testerink, Utrecht University, The Netherlands
 *
 */
public class AnswerSets {
	public Prolog p;
	public BuiltIn bi;

	public AnswerSets(BuiltIn bi){
		this.p = bi.p;
		this.bi = bi;

	}

	public int collectorID=0;
	public PrimIntKeyHashMap<PrimIntArrayList> collections = new PrimIntKeyHashMap<PrimIntArrayList>(PrimIntKeyHashMap.OBJ);
	/**
	 * Collect results from a collection call.
	 * @param type Collector type (findall/bagof/setof).
	 * @param call The call.
	 * @return Whether the collection call was successful.
	 */
	public boolean collect(int type,int[] call){ // findall(X,Y,Z) (or setof or bagof)
		int nr_called = p.history.intdata[p.history.pointer-2];
		if(nr_called == 0){
			makeCollector();
			int[] x = makeCollectGoal(call, type);
			
			PrimIntArrayList goals = new PrimIntArrayList(false);
			PrimIntArrayList y = new PrimIntArrayList(true);
			int startY = p.harvester.scanElement(call, 3, false, false);
			
			goals.addAr(p.extractArray(call, startY, p.harvester.scanElement(call, startY, true, false)));
			goals.addAr(x);
			p.addSubGoals(goals);
		} else if(nr_called == 1){
			int id = p.history.intdata[p.history.pointer-4];
			if(p.trace)System.out.println(": collecting reapings from collector "+id+" and building a list from it");
			PrimIntArrayList reapings = collections.getObj(id);
			PrimIntArrayList list = new PrimIntArrayList(true);
			list.addInt(p.harvester.LIST); // the first list items
			list.addInt(reapings.pointer);
			for(int j = 0; j < reapings.pointer; j++){ // add the list population
				int[] prim = reapings.getAr(j);
				//p.localVarMap.clear();
				p.harvester.scanElement(prim, 0, false, false); // scan over and replace variable numbers
				for(int i : prim) list.addInt(i); // add the element to the list
			}
			if(reapings.pointer>0){ // add tail
				list.addInt(p.harvester.LIST);
				list.addInt(0);
			}
			reapings = null;
			int cursor = 3;
			cursor = p.harvester.scanElement(call, cursor, false, false); // skip X
			cursor = p.harvester.scanElement(call, cursor, false, false); // skip Y
			return p.harvester.unifyTerm(call, p.listToPrim(list), cursor, 0,false,false); // unify Z with the list
		} else {
			if(p.trace)System.out.println("Can't redo a collection predicate.");
			return false; 
		}
		return true;
	}
	/**
	 * Perform measures to enable the collection of answers.
	 */
	public void makeCollector(){
		while(collections.getObj(collectorID)!=null)
			collectorID = (collectorID+1)%Integer.MAX_VALUE;
		collections.putObj(collectorID, new PrimIntArrayList(false));
		p.history.addInt(p.history.pointer-2,collectorID);// shove moverulecursor 1 to the right
		p.history.addInt(p.history.pointer-2,p.COLLECTORID);// shove moverulecursor 1 to the right
		if(p.trace)System.out.println(": making a collector with id "+collectorID);
	}
	/**
	 * Make an artificial goal to support collection.
	 * @param call Call of the collection predicate.
	 * @param type Type of collector.
	 * @return Goal to be added on the todo stack of the resolution algorithm.
	 */
	public int[] makeCollectGoal(int[] call, int type){
		int[] xcopy = null; // als X een variabele is, dan moet je z'n value hebben
		if(call[3]==IntHarvester.VARIABLE){
			int[] xvalue = p.substitution.value(call[4]);
			if(xvalue != null)
				xcopy = p.extractArray(xvalue, 0, p.harvester.scanElement(xvalue, 0, true, false));
			else {
				xcopy = new int[]{IntHarvester.VARIABLE,call[4]};
			}
		} else xcopy = p.extractArray(call, 3, p.harvester.scanElement(call, 3, true, false));
		int[] x = new int[6+xcopy.length];
		x[0] = p.harvester.PREDICATE;
		x[1] = bi.IC_BUILT_IN_COLLECTOR_DO_NOT_USE;
		x[2] = 2;
		x[3] = p.harvester.NUMBER;
		x[4] = collectorID;
		x[5] = type;
		for(int i = 0; i < xcopy.length; i++){
			x[6+i] = xcopy[i];
		}
		return x;
	}
	
	/**
	 * Add a result to the collection.
	 * @param call Result to be added.
	 */
	public void expandCollection(int[] call){
		int type = call[5];
		PrimIntArrayList expansion = p.harvester.deepcopy(call, 6); 
		if(p.trace)System.out.println(": adding "+p.harvester.arToString(p.listToPrim(expansion),true)+" to the collection of collector "+call[4]);
		if(type == bi.IC_FINDALL || type == bi.IC_BAGOF) collections.getObj(call[4]).addAr(p.listToPrim(expansion)); 
	}
}
