package aplprolog.prolog.builtins;
import aplprolog.prolog.*;
/**
 * Built-ins for atom related functionalities. 
 * @author Bas Testerink, Utrecht University, The Netherlands
 *
 */
public class AtomBuilding {
	public Prolog p;
	public Arithmetics arithmetics;
	
	public AtomBuilding(BuiltIn bi){
		p = bi.p;
		arithmetics = bi.arithmetics;
	}
	
	/**
	 * Check whether the element is atomic.
	 * @param call Call with element.
	 * @param cursor Index of the element.
	 * @param parenthesized Whether we are inside parenthesized content.
	 * @return Whether the element is atomic.
	 */
	public boolean isAtom(int[] call, int cursor,boolean parenthesized){
		if(call == null) return false;
		if(call[cursor]==IntHarvester.PREDICATE)
			return call[cursor+2] == 0;
		if(call[cursor]==IntHarvester.VARIABLE)
			return isAtom(p.substitution.value(call[cursor+1]),0,parenthesized);
		if(call[cursor]==IntHarvester.LIST)
			return call[cursor+1]==0;
		if(call[cursor]==IntHarvester.CUT) return true;
		if(call[cursor]==IntHarvester.PARENTHESIZED&&call[cursor+1]==1&&!parenthesized)
			return isAtom(call,cursor+2,true);
		return false;
	}
	/**
	 * Concatenate two atoms.
	 * @param call Call of the predicate.
	 * @return Whether concatenation was successful.
	 */
	public boolean concatAtom(int[] call){  
		//String s = concatAtomList(call,3);
		int idA = call[4];
		if(call[3] == IntHarvester.VARIABLE) idA = p.substitution.value(idA)[1];
		else if(call[3] != IntHarvester.PREDICATE) return false;
		int c = p.harvester.scanElement(call, 3, false, false);
		int idB = call[c+1];
		if(call[c] == IntHarvester.VARIABLE) idB = p.substitution.value(idB)[1];
		else if(call[c] != IntHarvester.PREDICATE) return false; 
		String sA = p.strStorage.getString(idA).replaceAll("'", ""); 
		if(sA == null) return false;
		String sB = p.strStorage.getString(idB).replaceAll("'", "");
		if(sB == null) return false;
		String s = sA+sB;
		//if(s==null) return false;
		if(p.strStorage.getInt(s)==null)p.strStorage.add(s);
		int cursor = p.harvester.scanElement(call, p.harvester.scanElement(call, 3, false, false), false, false);
		return p.harvester.unifyTerm(call, new int[]{IntHarvester.PREDICATE,p.strStorage.getInt(s),0}, cursor, 0,false,false);
	}
	
	/**
	 * Concatenate two atoms.
	 * @param call Call of the predicate.
	 * @param cursor Position in the call.
	 * @return Concatenated string.
	 */
	public String concatAtomList(int[] call, int cursor){
		if(call == null) return "";
		if(call[cursor]==IntHarvester.VARIABLE) return concatAtomList(p.substitution.value(call[cursor+1]),0);
		if(call[cursor]!=IntHarvester.LIST) return null;
		String s = "";
		int nr = call[cursor+1];
		int id = -1;
		int c = cursor+2;
		for(int i = 0; i < nr; i++){
			if(call[c]==IntHarvester.PREDICATE){
				if(call[c+2]!=0) return null;
				id = call[c+1];
			} else if(call[c]==IntHarvester.VARIABLE){
				if(p.substitution.value(call[c+1])==null) return null;
				if(p.substitution.value(call[c+1])[0]!=IntHarvester.PREDICATE) return null;
				if(p.substitution.value(call[c+1])[2]!=0) return null;
				id = p.substitution.value(call[c+1])[1];
			} else return null;
			s += p.strStorage.getString(id).replaceAll("'", "");
			c = p.harvester.scanElement(call, c, false, false);
		}
		String s2 = "";
		if(nr>0) s2 = concatAtomList(call, c); // concat the tail
		if(s2 == null) return null;
		return s + s2;
	}
	
	/**
	 * Convert a number to an atom.
	 * @param call Call containing the number.
	 * @param cursor Start of the number.
	 * @return Whether conversion was successful.
	 */
	public boolean intToAtom(int[] call, int cursor){
		double d = arithmetics.evaluate(call);
		if(Double.isNaN(d)){
			if(p.trace)System.out.println(": first argument is not a number");
			return false;
		}
		int c2 = p.harvester.scanElement(call, cursor, false, false);
		String s = ""+d;
		if(p.strStorage.getInt(s)==null)p.strStorage.add(s);
		return p.harvester.unifyTerm(call, new int[]{IntHarvester.PREDICATE,p.strStorage.getInt(s),0}, c2, 0,false,false);
	}
}
