package oopl;

import java.io.FileInputStream;
import java.io.InputStream;
import java.io.StringReader;
import java.util.ArrayList;

import oopl.parsing.IntOOPL;
import aplprolog.prolog.Prolog;
import aplprolog.prolog.builtins.ExternalActions;

public class DistributedOOPL {
	public Prolog prolog = new Prolog();
	public static int DOOPL=0,OOPL=1,COMBI=2;
	ArrayList<OOPLEventListener> listeners = new ArrayList<OOPLEventListener>();
	private OOPLNetworkNode networkNode;
	private boolean possibly_regimented = false;
	private int type = 0;
	private String source,strType; // latest load call (for reload())
	private boolean isFile = false; // latest load call (for reload())
	
	public DistributedOOPL(){
		loadLibraries();
		networkNode = new OOPLNetworkNode(6661, false,this);
		networkNode.registerActions(prolog);
	}

//	public static void main(String[] args){
//		DistributedOOPL oopl = new DistributedOOPL();
//	}

	/**
	 * Handle an event.
	 * @param event The event in String representation.
	 * @param give_feedback Whether to wait for success.
	 * @return Whether the event could be handled successfully. 
	 */
	public boolean handleEvent(String event, boolean give_feedback){
		return handleEvent(prolog.mp.parseFact(event, prolog.strStorage, false),give_feedback);
	}

	/**
	 * Handle an event.
	 * @param event The event in array representation.
	 * @param give_feedback Whether to wait for success.
	 * @return Whether the event could be handled successfully. 
	 */
	public boolean handleEvent(int[] event, boolean give_feedback){
		int[] query = new int[event.length+7]; //@execution_cycle ( event ) (event may not contain variables)
		query[0] = prolog.harvester.PREDICATE;
		query[1] = prolog.mp.strID(type==DOOPL?"@execution_cycle":"@execution_cycle_oopl");
		query[2] = 1;
		for(int i = 0; i < event.length; i++) 
			query[i+3] = event[i];
		query[event.length+3] = 0;
		query[event.length+4] = event.length+3;
		query[event.length+5] = 0;
		query[event.length+6] = 2;
		prolog.query(query);
		for(OOPLEventListener listener : listeners) listener.handleEvent(event, true); // notify listeners
		return true;
	}
	
	/**
	 * Load a norm file (unloads previous files and resets the Prolog engine).
	 * @param file File name.
	 */
	public void load(String string, boolean isFile, String type){
		try{
			IntOOPL parser;
			if(isFile) parser = new IntOOPL(new FileInputStream(string));
			else parser = new IntOOPL(new StringReader(string));
			parser.parseOOPLConfiguration(prolog.strStorage,type);
			this.type = type.equals("DOOPL")?DOOPL:(type.equals("OOPL")?OOPL:COMBI);
			loadLibraries(); // reload the libraries
			prolog.takeOverRules(parser.pti.rules,false); // take over 2OPL base
			possibly_regimented = parser.oti.possiblyRegimented; // check if viol_bot occurs in a counts-as rule, if not: you don't need to keep track of asserts/retractions
			prolog.builtin.rulemods.logModifications = possibly_regimented; // log all retracts/asserts
			this.source = string;
			this.strType = type;
			this.isFile = isFile;
		} catch(Exception e){
			e.printStackTrace();
		}
	}
	
	/**
	 * Redo the latest load/3 call.
	 */
	public void reload(){
		load(source,isFile,strType);
	}

	/**
	 * Load the various norm libraries.
	 */
	public void loadLibraries(){
		prolog.rulebase.clear(); // clear the rules
		prolog.loadfile("StdLibrary.pl",true); // reload the library
		prolog.loadfile("DistributedNormLibrary.pl",true); // reload the library
	}
	
	/**
	 * Add a 2OPL listener.
	 * @param Listener to add
	 */
	public void addOOPLEventListener(OOPLEventListener e){
		listeners.add(e);
	}
	
	/**
	 * Remove a 2OPL listener.
	 * @param e Listener to remove.
	 */
	public void removeOOPLEventListener(OOPLEventListener e){
		listeners.remove(e);
	}

	/**
	 * Get the Prolog engine.
	 * @return The prolog engine.
	 */
	public Prolog getProlog(){
		return prolog;
	}

	/**
	 * Get the network module.
	 * @return The network module.
	 */
	public OOPLNetworkNode getNetworkNode(){
		return networkNode;
	}
}
