package oopl.GUI;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.FlowLayout; 
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener; 
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;

import javax.swing.JButton; 
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel; 
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.event.CaretEvent;
import javax.swing.event.CaretListener;

import oopl.DistributedOOPL;
import oopl.OOPL;

public class Editor extends JPanel implements ActionListener, CaretListener{
	JTextArea textarea = new JTextArea();
	JScrollPane jsp = new JScrollPane(textarea,JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
	JButton jbtSave, jbtApply, jbtApplyAndSave, jbtSaveAs, jbtOpenFile;
	String lastPath = null;
	JLabel position = new JLabel("Line: 0 Column: 0");
	JFileChooser jfc = new JFileChooser("./");
	String type = "DOOPL";

	DistributedOOPL oopl;
    private PrologState prologstate = null;

	public Editor(DistributedOOPL oopl, String file, String type){
		super(new BorderLayout());
		setBackground(Color.white);
		this.oopl = oopl;
		this.type = type;
		textarea.addCaretListener(this);
		add(jsp,BorderLayout.CENTER); 

		// Buttons
		jbtOpenFile = new JButton("Open 2OPL file");
		jbtSave = new JButton("Save");
		jbtApply = new JButton("Apply");
		jbtApplyAndSave = new JButton("Save & Apply");
		jbtSaveAs = new JButton("Save As");
		jbtOpenFile.addActionListener(this);
		jbtSave.addActionListener(this);
		jbtApply.addActionListener(this);
		jbtApplyAndSave.addActionListener(this);
		jbtSaveAs.addActionListener(this);

		JPanel buttons = new JPanel(new FlowLayout());
		buttons.add(jbtOpenFile);
		buttons.add(jbtSave);
		buttons.add(jbtApply);
		buttons.add(jbtApplyAndSave);
		buttons.add(jbtSaveAs);
		buttons.setMinimumSize(new Dimension(30, 30));
		buttons.add(position);
		add(buttons,BorderLayout.SOUTH);

		if(file!=null)readFile(new File(file));
	}

	public void actionPerformed(ActionEvent e) { 
		if(e.getSource() == jbtOpenFile){
			if(jfc.showOpenDialog(null) == jfc.APPROVE_OPTION)
				readFile(jfc.getSelectedFile());
		}else if(e.getSource() == jbtSave){
			if(lastPath == null) saveAs();
			else saveFile();
		} else if(e.getSource() == jbtSaveAs) saveAs();
		else if(e.getSource() == jbtApply) apply();
		else if(e.getSource() == jbtApplyAndSave){
			if(lastPath == null) saveAs();
			else saveFile();
			apply();
		}
	}

	private void readFile(File file){
		String path = file.getPath();
		if(path.length()<6||path.lastIndexOf(".2opl")!=path.length()-5) // must be a .2opl file
			JOptionPane.showMessageDialog(this, "Not a 2OPL file", "Open 2OPL file", JOptionPane.ERROR_MESSAGE);
		else {
			lastPath = path;
			try{
				BufferedReader br = new BufferedReader(new FileReader(file));
				textarea.setText("");
				String line = "";
				while((line = br.readLine()) != null){
					line = line.replaceAll("\\n", ""); // strip away line endings
					line = line.replaceAll("\\r", "");
					textarea.append(line+"\r\n");
				}
				br.close();
			} catch(Exception e){ System.out.println(e); }
		}
	}

	public void saveAs(){
		int choice = jfc.showSaveDialog(null);
		if(choice == jfc.APPROVE_OPTION){
			File file = jfc.getSelectedFile(); // get the goal file
			lastPath = file.getPath();
			try{
				file.createNewFile(); // creates new file if needed
				saveFile();
			} catch(Exception e){ System.out.println(e); }
		}
	}

	public void saveFile(){
		try{
			File file = new File(lastPath);
			file.createNewFile(); // creates new file if needed
			BufferedWriter bw = new BufferedWriter(new FileWriter(file)); 
			for(int lnr = 0; lnr < textarea.getLineCount(); lnr++){
				int length = textarea.getLineEndOffset(lnr)-textarea.getLineStartOffset(lnr);
				String line = textarea.getText(textarea.getLineStartOffset(lnr), length);
				line = line.replaceAll("\\n", ""); // strip away line endings
				line = line.replaceAll("\\r", "");
				bw.write(line+(lnr==textarea.getLineCount()-1?"":"\r\n"));
			}
			bw.flush();
			bw.close();
		} catch(Exception e){ System.out.println(e); }
	}

	public void caretUpdate(CaretEvent e) {
		try{
			int pos = textarea.getCaretPosition();
			int line = textarea.getLineOfOffset(pos); 
			int column = pos - textarea.getLineStartOffset(line);
			updatePosition(line,column);
		} catch(Exception ex){System.out.println(ex);}
	}

	private void updatePosition(int l, int c){ 
		position.setText("Line: "+l+" Column: "+c);
	}
	
	public void setPrologState(PrologState ps){
		prologstate = ps;
	}
	
	public void apply(){
		oopl.load(textarea.getText(),false, type);
		// update prolog state GUI
		if(prologstate != null) prologstate.status.setText(oopl.getProlog().toString());
	}
}
