package oopl.GUI; 

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

import oopl.DistributedOOPL;
import oopl.OOPL;

import java.net.InetAddress;
import java.net.NetworkInterface;
import java.util.*; 

/**
 * GUI for 2OPL.
 * @author Bas Testerink, Utrecht University, The Netherlands
 */
public class GUI extends JFrame implements ActionListener { 
    private final static String TITLE = "2OPL Interpreter";
    private final static int WIDTH = 800, HEIGHT = 600;
    
    // GUI elementen
    private JMenuBar jMenuBar;
    private JMenu jmnuHelp, jmnuFile;
    private JMenuItem jmiHelp,jmiAbout,jmiOpenFile,jmiReloadFile,jmiExit; 
    private JTabbedPane jTabbedPane;
    private Editor editor;
    private PrologState prologState;
    private Monitor monitor;
    private Model model;
    private Network network;
    private JFileChooser jfc = new JFileChooser("./");
    
    //private OOPL oopl;
    private DistributedOOPL oopl;
    private String type = "DOOPL";
    
    /**
     * Create new GUI.
     */
    public GUI(DistributedOOPL oopl, String orgFile, String type, JComponent envGUI, int port) {
    	this.oopl = oopl;
    	this.type = type;
    	if(orgFile!=null)oopl.load(orgFile,true,type);
    	
        // Menubar
        jMenuBar = new JMenuBar();
        setJMenuBar(jMenuBar);
        
        // File menu
        jmnuFile = new JMenu("File");
        jMenuBar.add(jmnuFile);

        jmiOpenFile = new JMenuItem("Open and load file");
        jmiOpenFile.addActionListener(this);
        jmnuFile.add(jmiOpenFile);
        jmiReloadFile = new JMenuItem("Reload file");
        jmiReloadFile.addActionListener(this);
        jmnuFile.add(jmiReloadFile);
        jmiExit = new JMenuItem("Exit");
        jmiExit.addActionListener(this);
        jmnuFile.add(jmiExit);
        
        // Help menu
        jmnuHelp = new JMenu("Help");
        jMenuBar.add(jmnuHelp);

        jmiHelp = new JMenuItem("Help");
        jmiHelp.addActionListener(this);
        jmnuHelp.add(jmiHelp);
        jmiAbout = new JMenuItem("About");
        jmiAbout.addActionListener(this);
        jmnuHelp.add(jmiAbout);
        
        // Tabbed pane
        jTabbedPane = new JTabbedPane();
        jTabbedPane.setBackground(Color.white);
        add(jTabbedPane, BorderLayout.CENTER);
        
        editor = new Editor(oopl,orgFile,type);
        monitor = new Monitor(oopl);
        prologState = new PrologState(oopl);
        network = new Network(oopl.getNetworkNode(), port);
        model = new Model(oopl, envGUI);
        editor.setPrologState(prologState);

        jTabbedPane.add("2OPL editor", editor);
        jTabbedPane.add("System logs", monitor);
        jTabbedPane.add("Prolog status", prologState);
        jTabbedPane.add("Network connection", network);
        jTabbedPane.add("Environment model", model);
        
        // Frame properties
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        setTitle(TITLE);
        setSize(WIDTH, HEIGHT);
        setLocationRelativeTo(null); 
        setVisible(true);
    }
    
    /**
     * ActionListener.
     * @param e ActionEvent.
     */
    public void actionPerformed(ActionEvent e) {
        if(e.getSource() == jmiExit) System.exit(0);
        else if(e.getSource() == jmiOpenFile){
        	int choice = jfc.showOpenDialog(null);
        	if(choice == jfc.APPROVE_OPTION){
        		System.out.println(jfc.getSelectedFile());
        	}
        }
    }  
    
    public Network getNetwork(){
    	return network;
    }
    
    public static void main(String[] args) throws Exception {
        // Make new GUI object
        new GUI(new DistributedOOPL(),"testorg.2opl","COMBI",null,6789);
    }
}
