package oopl.GUI;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;

import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JToggleButton;

import oopl.DistributedOOPL;
import oopl.OOPL;
import oopl.OOPLEventListener;

public class Monitor extends JPanel implements ActionListener, OOPLEventListener{
	JTextArea events = new JTextArea(); // for the event log
	JScrollPane jspEvents = new JScrollPane(events,JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
	JTextArea networklog = new JTextArea(); // for the event log
	JScrollPane jspNetwork = new JScrollPane(networklog,JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
	JTextArea prologlog = new JTextArea(); // for the event log
	JScrollPane jspProlog = new JScrollPane(prologlog,JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
	JButton jbtClearEvents, jbtClearNetwork, jbtClearProlog, jbtSaveEvents, jbtSaveNetwork, jbtSaveProlog, jbtTexReport, jbtToPDF, jbtClearAll, jbtEnableAll, jbtDisableAll;
	JToggleButton jtbtEnableEvents, jtbtEnableNetwork, jtbtEnableProlog;
	boolean logEvents=false,logNetwork=false,logProlog=false;
	TexBuilder texBuilder = new TexBuilder();
	
	ArrayList<String> prologEntries = new ArrayList<String>(); // Prolog.toString() results after each event
	ArrayList<String> eventEntries = new ArrayList<String>(); // table entries for events
	ArrayList<String> networkEntries = new ArrayList<String>(); // table entries for events
	
	DistributedOOPL oopl;
	
	public Monitor(DistributedOOPL oopl){
		super(new BorderLayout());
		setBackground(Color.white);
		this.oopl = oopl;
		oopl.addOOPLEventListener(this);
		
		// General buttons:
		JPanel generalButtons = new JPanel(new FlowLayout());
		jbtTexReport = new JButton("Tex report");
		jbtToPDF = new JButton("To PDF");
		jbtEnableAll = new JButton("Enable all");
		jbtDisableAll = new JButton("Disable all");
		jbtClearAll = new JButton("Clear all logs");
		jbtTexReport.addActionListener(this);
		jbtToPDF.addActionListener(this);
		jbtEnableAll.addActionListener(this);
		jbtDisableAll.addActionListener(this);
		jbtClearAll.addActionListener(this);
		generalButtons.add(new JLabel("<html><h2>General: </h2></html>"));
		generalButtons.add(jbtTexReport);
		generalButtons.add(jbtToPDF);
		generalButtons.add(jbtEnableAll);
		generalButtons.add(jbtDisableAll);
		generalButtons.add(jbtClearAll);
		generalButtons.setMinimumSize(new Dimension(30,30));
		
		//Logs:
		JPanel logs = new JPanel(new GridLayout(3, 1));
		logs.setBackground(Color.white);
		
		// Event log:
		JPanel eventButtons = new JPanel(new FlowLayout());
		jbtClearEvents = new JButton("Clear log");
		jbtSaveEvents = new JButton("Save log");
		jtbtEnableEvents = new JToggleButton("Enable logging: OFF");
		jbtClearEvents.addActionListener(this);
		jbtSaveEvents.addActionListener(this);
		jtbtEnableEvents.addActionListener(this);
		eventButtons.add(jtbtEnableEvents);
		eventButtons.add(jbtClearEvents);
		eventButtons.add(jbtSaveEvents);
		JPanel jpnlEvents = new JPanel(new BorderLayout());
		jpnlEvents.add(new JLabel("<html><h2>Event log:</h2></html>"),BorderLayout.NORTH);
		jpnlEvents.add(jspEvents,BorderLayout.CENTER);
		jpnlEvents.add(eventButtons,BorderLayout.SOUTH);

		// TCP/IP log:
		JPanel networkButtons = new JPanel(new FlowLayout());
		jbtClearNetwork = new JButton("Clear log");
		jbtSaveNetwork = new JButton("Save log");
		jtbtEnableNetwork = new JToggleButton("Enable logging: OFF");
		jbtClearNetwork.addActionListener(this);
		jbtSaveNetwork.addActionListener(this);
		jtbtEnableNetwork.addActionListener(this);
		networkButtons.add(jtbtEnableNetwork);
		networkButtons.add(jbtClearNetwork);
		networkButtons.add(jbtSaveNetwork);
		JPanel jpnlNetwork = new JPanel(new BorderLayout());
		jpnlNetwork.add(new JLabel("<html><h2>Network log:</h2></html>"),BorderLayout.NORTH);
		jpnlNetwork.add(jspNetwork,BorderLayout.CENTER);
		jpnlNetwork.add(networkButtons,BorderLayout.SOUTH);

		// Prolog log:
		JPanel prologButtons = new JPanel(new FlowLayout());
		jbtClearProlog = new JButton("Clear log");
		jbtSaveProlog = new JButton("Save log");
		jtbtEnableProlog = new JToggleButton("Enable logging: OFF");
		jbtClearProlog.addActionListener(this);
		jbtSaveProlog.addActionListener(this);
		jtbtEnableProlog.addActionListener(this);
		prologButtons.add(jtbtEnableProlog);
		prologButtons.add(jbtClearProlog);
		prologButtons.add(jbtSaveProlog);
		JPanel jpnlProlog = new JPanel(new BorderLayout());
		jpnlProlog.add(new JLabel("<html><h2>Prolog log:</h2></html>"),BorderLayout.NORTH);
		jpnlProlog.add(jspProlog,BorderLayout.CENTER);
		jpnlProlog.add(prologButtons,BorderLayout.SOUTH);

		logs.add(jpnlEvents);
		logs.add(jpnlNetwork);
		logs.add(jpnlProlog);
		
		add(generalButtons,BorderLayout.NORTH);
		add(logs,BorderLayout.CENTER);
	}
 
	public void actionPerformed(ActionEvent e) { 
		if(e.getSource() == jbtClearAll){
			events.setText("");eventEntries.clear();
			networklog.setText("");networkEntries.clear();
			prologlog.setText("");prologEntries.clear();
		} else if(e.getSource() == jbtClearEvents){
			events.setText("");eventEntries.clear();
		} else if(e.getSource() == jbtClearNetwork){
			networklog.setText("");networkEntries.clear();
		} else if(e.getSource() == jbtClearProlog){
			prologlog.setText("");prologEntries.clear();
		} else if(e.getSource() == jbtDisableAll){
			jtbtEnableEvents.setSelected(false); flickEvents();
			jtbtEnableNetwork.setSelected(false); flickNetwork();
			jtbtEnableProlog.setSelected(false); flickProlog();
		} else if(e.getSource() == jbtEnableAll){
			jtbtEnableEvents.setSelected(true); flickEvents();
			jtbtEnableNetwork.setSelected(true); flickNetwork();
			jtbtEnableProlog.setSelected(true); flickProlog();
		} else if(e.getSource() == jtbtEnableEvents)flickEvents();
		else if(e.getSource() == jtbtEnableNetwork)flickNetwork();
		else if(e.getSource() == jtbtEnableProlog)flickProlog();
		else if(e.getSource() == jbtSaveEvents){
			
		} else if(e.getSource() == jbtSaveNetwork){
			
		} else if(e.getSource() == jbtSaveProlog){
			
		} else if(e.getSource() == jbtTexReport){
			texBuilder.generateLog(this);
		} else if(e.getSource() == jbtToPDF){
			texBuilder.generateLog(this);
			File f = new File("./logs/log.toc"); f.delete();
			f = new File("./logs/log.aux"); f.delete();
			f = new File("./logs/log.log"); f.delete();
			f = new File("./logs/log.out"); f.delete();
			f = new File("./logs/log.pdf"); f.delete();
			texBuilder.commando("pdflatex ./logs/log.tex -aux-directory=./logs/");
			texBuilder.commando("pdflatex ./logs/log.tex -aux-directory=./logs/");
			texBuilder.commando("log.pdf");
		}
	}
	
	private void flickEvents(){
		boolean on = jtbtEnableEvents.isSelected();
		if(on){
			jtbtEnableEvents.setText("Enable logging:  ON");
			logEvents = true;
		} else {
			jtbtEnableEvents.setText("Enable logging: OFF");
			logEvents = false;
		}
	}
	
	private void flickNetwork(){
		boolean on = jtbtEnableNetwork.isSelected();
		if(on){
			jtbtEnableNetwork.setText("Enable logging:  ON");
			logNetwork = true;
		} else {
			jtbtEnableNetwork.setText("Enable logging: OFF");
			logNetwork = false;
		}
	}
	
	private void flickProlog(){
		boolean on = jtbtEnableEvents.isSelected();
		if(on){
			jtbtEnableProlog.setText("Enable logging:  ON");
			logProlog = true;
		} else {
			jtbtEnableProlog.setText("Enable logging: OFF");
			logProlog = false;
		}
	} 
 
	public void handleEvent(int[] event, boolean result) { 
		if(logEvents||logProlog){ // if you are actually monitoring events
			String str = oopl.getProlog().harvester.elemToString(event, 0, false); // get the string representation
			Calendar cal = Calendar.getInstance();
			SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
			String date = sdf.format(cal.getTime());
			if(logEvents){
				eventEntries.add(texBuilder.tablerow(""+eventEntries.size(), str, ""+result, date)+"\r\n");
				events.append(date+": "+str+ " (was"+(result?" not ":" ")+"regimented)\r\n");
			}
			if(logProlog){
				prologEntries.add("\\subsection{Prolog state "+prologEntries.size()+"}\r\n"+
						texBuilder.prologBeginTable+
						texBuilder.tablerow(""+prologEntries.size(), str, ""+result, date)+
						texBuilder.prologEndTable+
						"\r\n"+
						"\\lstset{breaklines=true}\\begin{lstlisting}[language=Prolog, numbers=left]\r\n"+oopl.getProlog().toString()+"\\end{lstlisting}\r\n");
				prologlog.append(date+": Effects of "+str+" are processed.\r\n");
			}
		}
	}
}
