package oopl.GUI;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.ArrayList;

import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.JToggleButton;
import javax.swing.event.CaretEvent;
import javax.swing.event.CaretListener;

import oopl.DistributedOOPL;
import oopl.OOPL;
import oopl.OOPLEventListener;

public class PrologState extends JPanel implements ActionListener, CaretListener, KeyListener, OOPLEventListener{
	JSplitPane pane;
	JTextArea status = new JTextArea(); // for the prolog status
	JTextArea returns = new JTextArea(); // for the prolog return values
	JScrollPane jsp = new JScrollPane(status,JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
	JPanel consolePanel, queryPanel, returnPanel, buttonPanel;
	JTextField queryField = new JTextField();
	JButton jbtSave,jbtSaveAs,jbtOpenFile;
	JToggleButton jtbtKeepUp;
	JLabel position = new JLabel("Line: 0 Column: 0");
	String lastPath = null;
	JFileChooser jfc = new JFileChooser("./");
	DistributedOOPL oopl;
	ArrayList<String> history = new ArrayList<String>();// history of queries
	int historyC = 0;
	boolean keepUp = false;
	
	public PrologState(DistributedOOPL oopl){
		super(new BorderLayout());
		setBackground(Color.white); 
		this.oopl = oopl;
		oopl.addOOPLEventListener(this);
		
		// The query bar
		GridBagLayout gl = new GridBagLayout();
		GridBagConstraints fieldConstraints = new GridBagConstraints(); // some constraints for good resizing
		fieldConstraints.fill = GridBagConstraints.HORIZONTAL;
		fieldConstraints.anchor = GridBagConstraints.NORTHWEST;
		fieldConstraints.weightx = 1.0;
		fieldConstraints.gridwidth = GridBagConstraints.REMAINDER;
		fieldConstraints.insets = new Insets(1, 1, 1, 1);
		GridBagConstraints labelConstraints = (GridBagConstraints) fieldConstraints.clone();
		labelConstraints.weightx = 0.0;
		labelConstraints.gridwidth = 1;
		queryPanel = new JPanel(gl);
		queryPanel.add(new JLabel("Query: "),labelConstraints);
		queryPanel.add(queryField,fieldConstraints);
		queryPanel.setMinimumSize(new Dimension(30, 30));

		// Buttons
		jbtOpenFile = new JButton("Open Prolog file");
		jbtSave = new JButton("Save & Apply");
		jbtSaveAs = new JButton("Save As");
		jtbtKeepUp = new JToggleButton("Keep updated: OFF");
		jbtOpenFile.addActionListener(this);
		jbtSave.addActionListener(this); 
		jbtSaveAs.addActionListener(this);
		jtbtKeepUp.addActionListener(this);

		JPanel buttons = new JPanel(new FlowLayout());
		buttons.add(jbtOpenFile);
		buttons.add(jbtSave);
		buttons.add(jbtSaveAs);
		buttons.add(jtbtKeepUp);
		buttons.add(position);
		buttons.setMinimumSize(new Dimension(30, 30));

		// Lower pane part
		consolePanel = new JPanel(new BorderLayout());
		consolePanel.add(buttons,BorderLayout.NORTH);
		consolePanel.add(new JScrollPane(returns,JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED),BorderLayout.CENTER);
		consolePanel.add(queryPanel,BorderLayout.SOUTH);

		// wrap
		status.addCaretListener(this);
		pane = new JSplitPane(JSplitPane.VERTICAL_SPLIT,jsp,consolePanel);
		add(pane,BorderLayout.CENTER);
		jsp.setMinimumSize(new Dimension(10, 300)); 
		
		// load the prolog state
		status.setText(oopl.getProlog().toString());
		queryField.addKeyListener(this);
	}

	public void actionPerformed(ActionEvent e) { 
		if(e.getSource() == jbtOpenFile) getFileText();
		else if(e.getSource() == jbtSave){
			if(lastPath == null) saveAs();
			else saveFile();
		} else if(e.getSource() == jbtSaveAs) saveAs();
		else if(e.getSource() == jtbtKeepUp){
			jtbtKeepUp.setText(jtbtKeepUp.isSelected()?"Keep updated:  ON":"Keep updated: OFF");
			keepUp = jtbtKeepUp.isSelected();
			if(keepUp) status.setText(oopl.getProlog().toString());
		}
	}

	public void getFileText(){
		int choice = jfc.showOpenDialog(null);
		if(choice == jfc.APPROVE_OPTION){
			File file = jfc.getSelectedFile();
			String path = file.getPath();
			if(path.length()<4||path.lastIndexOf(".pl")!=path.length()-3) // must be a .2opl file
				JOptionPane.showMessageDialog(this, "Not a Prolog file", "Open Prolog file", JOptionPane.ERROR_MESSAGE);
			else {
				lastPath = path;
				try{
					BufferedReader br = new BufferedReader(new FileReader(file));
					status.setText("");
					String line = "";
					while((line = br.readLine()) != null){
						line = line.replaceAll("\\n", ""); // strip away line endings
						line = line.replaceAll("\\r", "");
						status.append(line+"\r\n");
					}
					br.close();
				} catch(Exception e){ System.out.println(e); }
			}
		}
	}

	public void saveAs(){
		int choice = jfc.showSaveDialog(null);
		if(choice == jfc.APPROVE_OPTION){
			File file = jfc.getSelectedFile(); // get the goal file
			lastPath = file.getPath();
			try{
				file.createNewFile(); // creates new file if needed
				saveFile();
			} catch(Exception e){ System.out.println(e); }
		}
	}

	public void saveFile(){
		try{
			File file = new File(lastPath);
			file.createNewFile(); // creates new file if needed
			BufferedWriter bw = new BufferedWriter(new FileWriter(file)); 
			for(int lnr = 0; lnr < status.getLineCount(); lnr++){
				int length = status.getLineEndOffset(lnr)-status.getLineStartOffset(lnr);
				String line = status.getText(status.getLineStartOffset(lnr), length);
				line = line.replaceAll("\\n", ""); // strip away line endings
				line = line.replaceAll("\\r", "");
				bw.write(line+(lnr==status.getLineCount()-1?"":"\r\n"));
			}
			bw.flush();
			bw.close();
		} catch(Exception e){ System.out.println(e); }
	}

	public void caretUpdate(CaretEvent e) {
		try{
			int pos = status.getCaretPosition();
			int line = status.getLineOfOffset(pos); 
			int column = pos - status.getLineStartOffset(line);
			updatePosition(line,column);
		} catch(Exception ex){System.out.println(ex);}
	}

	private void updatePosition(int l, int c){ 
		position.setText("Line: "+l+" Column: "+c);
	}
 
	public void keyTyped(KeyEvent e) { 
	}
 
	public void keyPressed(KeyEvent e) { 
		
	}
 
	public void keyReleased(KeyEvent e) { 
		if(e.getKeyCode() == KeyEvent.VK_ENTER){
			String q = queryField.getText();
			if(!q.endsWith(".")) q = q + ".";
			returns.append("Query: "+q+"\r\nAnswer: \r\n");
			if(historyC!=history.size())history.remove(historyC);
			history.add(q); 
			historyC = history.size();
			queryField.setText("");
			oopl.getProlog().traceSol = true;
			oopl.getProlog().query(q);
			oopl.getProlog().traceSol = false;
			returns.append(oopl.getProlog().solStr);
			if(!returns.getText().endsWith("\r\n"))returns.append("\r\n");
			returns.append("###########\r\n");
		} else if(e.getKeyCode() == KeyEvent.VK_UP){
			historyC = Math.max(historyC-1, 0);
			queryField.setText(history.get(historyC));
		} else if(e.getKeyCode() == KeyEvent.VK_DOWN){
			historyC = Math.min(historyC+1,history.size());
			queryField.setText((historyC==history.size()?"":history.get(historyC)));
		} 
		
	}

	public void handleEvent(int[] event, boolean result) {
		if(result&&keepUp){
			status.setText(oopl.getProlog().toString());
		}
	}
}
