package oopl;
import java.io.FileInputStream;
import java.io.StringReader;
import java.util.ArrayList;

import oopl.parsing.IntOOPL; 
import aplprolog.Parsing.ParseException;
import aplprolog.datastructures.PrimIntArrayList;
import aplprolog.prolog.IntHarvester;
import aplprolog.prolog.Prolog;
import aplprolog.prolog.StringStorage;
import aplprolog.prolog.builtins.ExternalActions;
import aplprolog.prolog.builtins.ExternalTool;

/*
 * Phi_l: <Phi_c,M(Phi_x),Phi_d> ==> instantiate_norms :- Phi_c,uniqueassert(instantiation(M,Phi_l,Phi_x,Phi_d)),fail.
 * check_norms :- instantiation(obligation,Phi_l,Phi_x,Phi_d),Phi_x,assertunique(obey_norm(Phi_l)),retract(instantiation(obligation,Phi_l,Phi_x,Phi_d)),fail.
 * check_norms :- instantiation(obligation,Phi_l,Phi_x,Phi_d),Phi_d,assertunique(viol_norm(Phi_l)),retract(instantiation(obligation,Phi_l,Phi_x,Phi_d)),fail.
 * check_norms :- instantiation(forbidden,Phi_l,Phi_x,Phi_d),Phi_x,assertunique(viol_norm(Phi_l)),retract(instantiation(forbidden,Phi_l,Phi_x,Phi_d)),fail.
 * check_norms :- instantiation(forbidden,Phi_l,Phi_x,Phi_d),Phi_d,assertunique(obey_norm(Phi_l)),retract(instantiation(forbidden,Phi_l,Phi_x,Phi_d)),fail.
 */
public class OOPL {
	private Prolog prolog = new Prolog();
	int[] countsasquery = null, sanctionquery = null, violbotquery = null, effect_base = null, effect_query = new int[]{10}, inst_norms_query = null, check_norms_query = null;
	boolean possibly_regimented = false; // whether norms can be regimented, if false, all norms are enforced
	ArrayList<OOPLEventListener> listeners = new ArrayList<OOPLEventListener>();
	public IntOOPL parser = new IntOOPL(new StringReader(""));
	private OOPLNetworkNode networkNode = null;
	
	public OOPL(){
		try {
			prolog.loadfile("StdLibrary.pl",true); // load the library
			countsasquery = prolog.mp.parseQuery("counts_as_rule.", prolog.strStorage);
			sanctionquery = prolog.mp.parseQuery("sanction_rule.", prolog.strStorage);
			violbotquery = prolog.mp.parseQuery("viol_bot.", prolog.strStorage);
			inst_norms_query = prolog.mp.parseQuery("instantiate_norms.", prolog.strStorage);
			check_norms_query = prolog.mp.parseQuery("check_norms.", prolog.strStorage);
			prolog.builtin.rulemods.logModifications = possibly_regimented; // log all retracts/asserts
			int[] base_query = prolog.mp.parseQuery("norm_effect_rule.", prolog.strStorage);
			effect_base = new int[]{base_query[0],base_query[1],1};
			prolog.builtin.external.registerAction("send", networkNode, ExternalActions.INTAR, ExternalActions.INTAR);
			prolog.builtin.external.registerAction("atomtest", networkNode, ExternalActions.INTAR, ExternalActions.INTAR); // @external(network,atomtest(atom),Atom),@external(network,listtest(list),List),@external(network,numtest(number),Num).
			prolog.builtin.external.registerAction("numtest", networkNode, ExternalActions.INTAR, ExternalActions.INTAR);
			prolog.builtin.external.registerAction("listtest", networkNode, ExternalActions.INTAR, ExternalActions.INTAR);
			prolog.builtin.external.registerAction("broadcast", networkNode, ExternalActions.INTAR, ExternalActions.INTAR);
			prolog.builtin.external.registerAction("substitutions", networkNode, ExternalActions.INTAR, ExternalActions.INTAR);
			prolog.builtin.external.registerAction("entailed", networkNode, ExternalActions.INTAR, ExternalActions.INTAR);
		} catch (Exception e) {e.printStackTrace();}
	}
	
	public static void main(String[] args) throws ParseException{
		OOPL o = new OOPL(); 
		//o.handleEvent("move(karl,left)"); 
		o.doSingleQuery("position(franz,X,Y).");
		o.handleEvent((new IntOOPL(new StringReader(""))).parseFact("move(franz,right).", o.prolog.strStorage)); 
		o.prolog.traceSol = true;
		o.doSingleQuery("position(franz,X,Y)."); 
	}
	
	/**
	 * Handle an event.
	 * @param event The event in String representation.
	 * @return Whether the event could be handled successfully. 
	 */
	public boolean handleEvent(int[] event){
		prolog.builtin.externalCallQueue.clear(); // clear external actions (might still hold some from last event if it had a regimented measure)
		prolog.builtin.rulemods.changeLog.clear(); // clear changelog
		buildEffectQuery(event); // build the event query  
		if(doSingleQuery(effect_query)) // seek and perform the effect rule, return failure if there is no effect rule 
			doSingleQuery(inst_norms_query);// instantiate all the temporal norms
			doSingleQuery(check_norms_query); // check all obligations/prohibitions
			if(applyCountsAs()){ // apply the counts-as rules return failure if viol_bot was triggered
				applySanctionRules(); // apply the sanctioning rules
				executeActionQueue(); // send all the messages and perform external actions
				for(OOPLEventListener listener : listeners) listener.handleEvent(event, true); // notify listeners
				return true; // event successfully handled
			} else undoModifications(); // regimented, undo rule modifications
		for(OOPLEventListener listener : listeners) listener.handleEvent(event, false); // notify listeners
		return false;
	}
	
	/**
	 * Build the query for effect-rule application. The event gets wrapped in norm_effect_rule().
	 * @param event The event to be wrapped.
	 */
	private void buildEffectQuery(int[] event){
		int length = event.length+4;// needs the extra integers for the head
		if(effect_query.length!=length)effect_query = new int[length]; // create new array if needed
		for(int i = 0; i < 3; i++)effect_query[i] = effect_base[i]; // copy base integers
		for(int i = 3; i < length-1; i++)effect_query[i] = event[i-3]; // copy event integers
		effect_query[length-1] = effect_query[length-2]+1; // move split counter one forward
		effect_query[length-2] = -3; // add zero split
		int nrSplits = effect_query[length-1];
		for(int i = 0; i < nrSplits; i++)effect_query[length-2-i] += 3; // splits all move 3 forward
		int varC = length-2-nrSplits; // var cursor
		int nrVars = effect_query[varC];
		for(int v = 0; v < nrVars; v++){ // move each variable 3 forward as well
			int amount = effect_query[varC];
			varC--; // move first index
			for(int index = 0; index < amount; index++){
				effect_query[varC] += 3;
				varC--; // move further
			}
		}
	}

	/**
	 * Perform a single query and trace the resolution engine.
	 * @param query Query (String format).
	 */
	public boolean doSingleQuery(String query){
		//prolog.trace = true; // uncomment to trace the resolution engine
		//prolog.traceSol = true;
		boolean r = prolog.query(query);
		prolog.traceSol = false;
		prolog.trace = false;
		return r;
	}
	/**
	 * Perform a single query and trace the resolution engine.
	 * @param query Query (integer format).
	 */
	public boolean doSingleQuery(int[] query){
		//prolog.trace = true; // uncomment to trace the resolution engine
		boolean r = prolog.query(query); 
		return r;
	}
	
	/**
	 * Load a 2OPL file.
	 * @param file File name.
	 */
	public void loadfile(String file){
		try{
			IntOOPL parser = new IntOOPL(new FileInputStream(file));
			parser.parseOOPLConfiguration(prolog.strStorage, "OOPL");
			prolog.rulebase.clear(); // clear the rules
			//prolog.loadfile("StdLibrary.pl",true); // reload the library
			prolog.loadfile("StdLibrary.pl",false); // load the library
			prolog.loadfile("DistributedNormLibrary.pl",false); // load the library
			prolog.takeOverRules(parser.pti.rules,false); // take over 2OPL base
			possibly_regimented = parser.oti.possiblyRegimented; // check if viol_bot occurs in a counts-as rule, if not: you don't need to keep track of asserts/retractions
			prolog.builtin.rulemods.logModifications = possibly_regimented; // log all retracts/asserts
		} catch(Exception e){
			e.printStackTrace();
		}
	}
	
	/**
	 * Load a 2OPL string.
	 * @param file File name.
	 */
	public void loadFromString(String text){
		try{
			IntOOPL parser = new IntOOPL(new StringReader(text));
			parser.parseOOPLConfiguration(prolog.strStorage, "OOPL");
			prolog.rulebase.clear(); // clear the rules
			prolog.loadfile("2oplLibrary.pl",false); // reload the library
			prolog.takeOverRules(parser.pti.rules,false); // take over 2OPL base
			possibly_regimented = parser.oti.possiblyRegimented; // check if viol_bot occurs in a counts-as rule, if not: you don't need to keep track of asserts/retractions
			prolog.builtin.rulemods.logModifications = possibly_regimented; // log all retracts/asserts
		} catch(Exception e){
			e.printStackTrace();
		}
	}

	/**
	 * Applies the counts-as rules. Returns false if viol_bot can be derived.
	 * @return Whether viol_bot can be derived in the closure of the counts-as rules.
	 */
	public boolean applyCountsAs(){ 
		prolog.rulesModified = true;
		while(prolog.rulesModified){
			doSingleQuery(countsasquery);  
			if(possibly_regimented){ // skip this step in event-only environments
				if(doSingleQuery(violbotquery)){ // norm is regimented
					return false; 
				}
			}
		}
		return true;
	}
	
	/**
	 * Apply the sanction rules.
	 */
	public void applySanctionRules(){
		prolog.rulesModified = true;
		while(prolog.rulesModified){
			doSingleQuery(sanctionquery);  
		}
	}
	
	/**
	 * Execute the actions that were stored in the Prolog engine.
	 */
	public void executeActionQueue(){
		PrimIntArrayList queue = prolog.builtin.externalCallQueue;
		for(int i = 0; i < queue.pointer; i++){ 
			boolean r = doSingleQuery(queue.arraydata[i]); 
		}
	}
	
	/**
	 * Undo all the rule modifications.
	 */
	public void undoModifications(){
		PrimIntArrayList stack = prolog.builtin.rulemods.changeLog;
		prolog.builtin.rulemods.logModifications = false;
		for(int i = stack.pointer-1; i >= 0; i--){ // undo in reverse order
			doSingleQuery(stack.arraydata[i]);  
		}
		prolog.builtin.rulemods.logModifications = true;
	}

	public void addOOPLEventListener(OOPLEventListener e){
		listeners.add(e);
	}
	public void removeOOPLEventListener(OOPLEventListener e){
		listeners.remove(e);
	}
	
	public Prolog getProlog(){
		return prolog;
	}
	
	public OOPLNetworkNode getNetworkNode(){
		return networkNode;
	}
}
