package oopl;

import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.StringReader;
import java.net.InetAddress;
import java.net.Socket;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;

import aplprolog.Parsing.IntProlog;
import aplprolog.prolog.IntHarvester;
import aplprolog.prolog.Prolog;
import aplprolog.prolog.builtins.ExternalActions;
import aplprolog.prolog.builtins.ExternalTool;

import networkConnection.ClientMessageHandler;
import networkConnection.NetworkListener;
import networkConnection.Server;


public class OOPLNetworkNode implements ClientMessageHandler, NetworkListener, ExternalTool{
	public Server server = null;
	public String name = "network";
	ArrayList<OOPLNetworkListener> listeners = new ArrayList<OOPLNetworkListener>();
	Calendar cal = Calendar.getInstance();
	SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
	DistributedOOPL oopl;

	public OOPLNetworkNode(int serverPort, boolean startServer, DistributedOOPL oopl){
		this.oopl = oopl;
		server = new Server(serverPort, this);
		server.addListener(this);
		if(startServer)server.start();
	}
	public static void main(String[] args){
		/*
		 * Server object moet gegeven een bytearray en doel string een message zenden en/of wachten op returnvalue.
		 * Vanuit 2OPL heb je norm instances die gecommuniceerd worden.
		 * Vanuit de road simulator heb je een centraal punt waaraan wegen zich kunnen koppelen. Daarna heb je messages
		 * die agent overdrachten aangeven. 
		 */
		DistributedOOPL oopl = new DistributedOOPL();
		OOPLNetworkNode o = new OOPLNetworkNode(6661,true,oopl);
		OOPLNetworkNode o2 = new OOPLNetworkNode(6662,false,oopl);
		System.out.println((String)o2.server.sendMessage("Dit is het bericht.", "192.168.1.109", 6661, true, true));
		try{Thread.sleep(200);}catch(Exception e){e.printStackTrace();}
		o.server.disconnectServer();
	}
	
	public void addNetworkListener(OOPLNetworkListener nl){
		listeners.add(nl);
	}
	public void removeNetworkListener(OOPLNetworkListener nl){
		listeners.remove(nl);
	}

	public void handle(Socket socket, ObjectInputStream ois, ObjectOutputStream oos, Object msg) {
		try{
			String message = (String)msg;
			notifyMessageTransfer(sdf.format(cal.getTime()), socket.getInetAddress().getHostAddress(), InetAddress.getLocalHost().getHostAddress(), message, true);
			String answer = "(default answer)";
			if(message.startsWith("broadcast")){ answer = null; oopl.prolog.query("@update_facts("+message.substring(10)+")."); }
			else if(message.startsWith("entailed")) answer = oopl.prolog.query(message.substring(9)+".")?"true":"false";
			else if(message.startsWith("substitutions")) answer = doSubstitutions(message.substring(14));
			if(answer!=null){
				oos.writeObject(answer);
				oos.flush();
				notifyMessageTransfer(sdf.format(cal.getTime()), InetAddress.getLocalHost().getHostAddress(), socket.getInetAddress().getHostAddress(), answer, false);
			}
		} catch(Exception ex){
			ex.printStackTrace();
		}
	}

	private String doSubstitutions(String q){
		String query = "findall("+q+","+q+",TheResultList_).";
		oopl.prolog.query(query);
		int varID = oopl.prolog.mp.vars.get("TheResultList_")+1;
		return oopl.prolog.harvester.elemToString(oopl.prolog.substitution.value(varID), 0, true);
	}
	
	public void notifyMessageTransfer(String date, String from, String to, Object message, boolean received){
		for(OOPLNetworkListener nl : listeners) nl.notifyMessageTransfer(date, from, to, (String)message, received);
	}
	
	public void notifyOverheadEvent(String date, String event) {
		for(OOPLNetworkListener nl : listeners) nl.notifyOverheadEvent(date, event);
	}

	public void registerActions(Prolog prolog){
		prolog.builtin.external.registerAction("broadcast", this, ExternalActions.INTAR, ExternalActions.INTAR);
		prolog.builtin.external.registerAction("substitutions", this, ExternalActions.INTAR, ExternalActions.INTAR);
		prolog.builtin.external.registerAction("entailed", this, ExternalActions.INTAR, ExternalActions.INTAR);
		prolog.builtin.external.registerAction("send", this, ExternalActions.INTAR, ExternalActions.INTAR);
		prolog.builtin.external.registerAction("atomtest", this, ExternalActions.INTAR, ExternalActions.INTAR); // @external(network,atomtest(atom),Atom),@external(network,listtest(list),List),@external(network,numtest(number),Num).
		prolog.builtin.external.registerAction("numtest", this, ExternalActions.INTAR, ExternalActions.INTAR);
		prolog.builtin.external.registerAction("listtest", this, ExternalActions.INTAR, ExternalActions.INTAR);
	}
	
	public void handleCall(int[] call, ExternalActions ea, int returnType) {
		if(call[1] == ea.bi.p.strStorage.getInt("atomtest")){
			String msg = ea.bi.p.strStorage.getString(call[4]);
			notifyMessageTransfer(sdf.format(cal.getTime()), "Prolog", "Network Node", "atomtest: "+msg,true);
			ea.intResult = new int[]{IntHarvester.PREDICATE,ea.bi.p.strStorage.getInt("true"),0};
		} else if(call[1] == ea.bi.p.strStorage.getInt("listtest")){
			String msg = ea.bi.p.strStorage.getString(call[4]);
			notifyMessageTransfer(sdf.format(cal.getTime()), "Prolog", "Network Node", "listtest: "+msg,true);
			ea.intResult = new int[]{IntHarvester.LIST,3,IntHarvester.PREDICATE,ea.bi.p.strStorage.getInt("true"),0,IntHarvester.PREDICATE,ea.bi.p.strStorage.getInt("false"),0,IntHarvester.NUMBER,283,23,IntHarvester.LIST,0};
		} else if(call[1] == ea.bi.p.strStorage.getInt("numtest")){
			String msg = ea.bi.p.strStorage.getString(call[4]);
			notifyMessageTransfer(sdf.format(cal.getTime()), "Prolog", "Network Node", "numtest: "+msg,true);
			ea.intResult = new int[]{IntHarvester.NUMBER,12,40};
		} else if(call[1] == ea.bi.p.strStorage.getInt("entailed")){//@external(network,entailed('192.168.1.109',6789,position(A,B,C) ),R).
			String msg = "entailed "+ea.bi.p.harvester.elemToString(call, 9, true);
			String returnMsg = (String)server.sendMessage(msg, host(call,ea), port(call,ea), true, true);
			//System.out.println(returnMsg);
			ea.intResult = new int[]{IntHarvester.PREDICATE,ea.bi.p.strStorage.getInt(returnMsg),0};
			notifyMessageTransfer(sdf.format(cal.getTime()), "Prolog", "Network Node", "entailed request",true);
		} else if(call[1] == ea.bi.p.strStorage.getInt("substitutions")){ //@external(network,substitutions('192.168.1.109',6789,position(A,B,C) ),R).
			String msg = "substitutions "+ea.bi.p.harvester.elemToString(call, 9, true);
			String returnMsg = (String)server.sendMessage(msg, host(call,ea), port(call,ea), true, true);
			IntProlog parser = new IntProlog(new StringReader(returnMsg));
			ea.intResult = parser.parseTerm(returnMsg, oopl.prolog.strStorage);
			notifyMessageTransfer(sdf.format(cal.getTime()), "Prolog", "Network Node", "substitutions request",true);
		} else if(call[1] == ea.bi.p.strStorage.getInt("broadcast")){//@broadcast([@lbl(id,plus(test(123))),@lbl(id,update(move(agent0,2,2)))]).
			String msg = "broadcast "+ea.bi.p.harvester.elemToString(call, 9, true);
			server.sendMessage(msg, host(call,ea), port(call,ea), false, true);
			ea.intResult = new int[]{IntHarvester.PREDICATE,ea.bi.p.strStorage.getInt("true"),0};
			notifyMessageTransfer(sdf.format(cal.getTime()), "Prolog", "Network Node", "broadcast request",true);
		} else if(call[1] == ea.bi.p.strStorage.getInt("send")){//@external(network,send('the message','192.168.1.109',6789),R)
			String host = ea.bi.p.strStorage.getString(call[7]);
			host = host.substring(1,host.length()-1);
			long l1 = ((long)call[10]<<32)>>>32;
			long l2 = ((long)call[11]<<32)>>>32;
			int port = (int)Double.longBitsToDouble((l1<<32)|l2);
			notifyMessageTransfer(sdf.format(cal.getTime()), "Prolog", "Network Node", "send: "+ea.bi.p.strStorage.getString(call[4])+" "+host+":"+port,true);
			String returnMsg = (String)server.sendMessage(ea.bi.p.strStorage.getString(call[4]), host, port, true, true);
			Integer i = ea.bi.p.strStorage.getInt(returnMsg);
			if(i == null) i = ea.bi.p.strStorage.add("\'"+returnMsg+"\'");
			ea.intResult = new int[]{IntHarvester.PREDICATE,i,0};
		}
	}

	private String host(int[] call, ExternalActions ea){
		String host = ea.bi.p.strStorage.getString(call[4]);
		return host.substring(1,host.length()-1);
	}
	private int port(int[] call, ExternalActions ea){
		long l1 = ((long)call[7]<<32)>>>32;
		long l2 = ((long)call[8]<<32)>>>32;
		return (int)Double.longBitsToDouble((l1<<32)|l2);
	}
	
	public void handleCall(Object[] call, ExternalActions ea, int returnType) {
	}
	public String getName(){
		return name;
	}
}
