package oopl.parsing;

import java.util.ArrayList;
import java.util.HashMap;

import aplprolog.datastructures.PrimIntArrayList;
import aplprolog.datastructures.PrimIntKeyHashMap;
import aplprolog.prolog.IntHarvester;
import aplprolog.prolog.Prolog;
import aplprolog.prolog.StringStorage;

public class PrologToInts {
	public PrimIntKeyHashMap<PrimIntKeyHashMap<PrimIntArrayList>> rules;
	public PrimIntArrayList array, splits;
	public StringStorage strStor;
	public HashMap<Integer,PrimIntArrayList> varIndices;
	public HashMap<String,Integer> vars;

	public PrologToInts(StringStorage strStor){
		this.strStor = strStor==null?new StringStorage():strStor;
	}

	////////////////////////
	// PRELIMINARY STUFF ///
	////////////////////////
	public void new_base(){
		varIndices = new HashMap<Integer, PrimIntArrayList>();
		vars = new HashMap<String,Integer>();
		array  = new PrimIntArrayList(true);
		splits = new PrimIntArrayList(true);
		rules = new PrimIntKeyHashMap<PrimIntKeyHashMap<PrimIntArrayList>>(PrimIntKeyHashMap.OBJ);
	} 

	/**
	 * Adds the current array as a rule to the rule base.
	 */
	public void add_rule(){
		// space in rule base
		int nameI = array.intdata[1]; // name 
		int arityI = array.intdata[2]; // and arity
		PrimIntKeyHashMap<PrimIntArrayList> name = rules.getObj(nameI);
		if(name==null){ name = new PrimIntKeyHashMap<PrimIntArrayList>(PrimIntKeyHashMap.OBJ); rules.putObj(nameI,name); }
		PrimIntArrayList arity = name.getObj(arityI);
		if(arity==null){ arity = new PrimIntArrayList(false); name.putObj(arityI,arity);}

		// splits and var indices (overhead integers)
		addVarOverhead();
		addSplitsOverhead();

		// conversion arraylist -> primitive array
		arity.addAr(getPrimArray(0));

//		Prolog p = new Prolog();
//		p.strStorage = strStor;
//		IntHarvester ih = new IntHarvester(p);
//		System.out.println(arStr(arity.arraydata[arity.pointer-1]));
//		System.out.println(ih.ruleToString(arity.arraydata[arity.pointer-1]));
		// clean up
		clear_overhead();
	}
	
	public void clear_overhead(){
		array.clear();
		splits.clear();
		vars.clear(); 
		varIndices.clear();
	}
	

	  /**
	   * Get all the splits and variable indices from the converted array.
	   */
	  public void fix_overhead(){
		  splits.clear();
		  for(PrimIntArrayList pial : varIndices.values()) pial.clear();
		  int c = 0; 
		  while(c != array.pointer){
			  c = overheadScan(c);
			  splits.addInt(c);
		  }
	  }
	  
	  /**
	   * Skip element and keep track of variables.
	   * @param start Cursor in the global parsing array.
	   * @return End integer of the element.
	   */
	  public int overheadScan(int start){
		int type = array.intdata[start];
		if(type==IntHarvester.PREDICATE){
			int x = start+3; // start of args
			int nrargs = array.intdata[start+2]; 
			for(int i = 0; i < nrargs;i++) x = overheadScan(x); // scan over arguments
			return x;
		} else if(type==IntHarvester.VARIABLE){
			int id = array.intdata[start+1];
			varIndices.get(id).addInt(start+1);
			return start+2;
		} else if(type==IntHarvester.LIST){
			int x = start+2; // start of elems
			int nrElems = array.intdata[start+1]; 
			for(int i = 0; i < nrElems;i++) x = overheadScan(x); // scan over elems
			if(nrElems>0)x = overheadScan(x);// also skip tail
			return x; 
		} else if(type==IntHarvester.NUMBER){
			return start+3;
		} else if(type==IntHarvester.PARENTHESIZED){
			int x = start+2; // start of args
			int nrElems = array.intdata[start+1];
			for(int i = 0; i < nrElems;i++) x = overheadScan(x); // scan over arguments
			return x;
		} else if(type==IntHarvester.CUT){ 
			return start+2;
		} else {
			return -1;
		}
	  }

	/**
	 * Adds integers about variable positions in the array.
	 */
	public void addVarOverhead(){
		for(int var = vars.size()-1; var >= 0; var--){
			PrimIntArrayList indices = varIndices.get(var);
			for(int i = indices.pointer-1; i >= 0; i--) array.addInt(indices.intdata[i]);
			array.addInt(indices.pointer);
		}
		array.addInt(vars.size());
	}

	/**
	 * Adds integers about split positions in the array.
	 */
	public void addSplitsOverhead(){
		for(int split = splits.pointer-1; split >= 0; split--) 
			array.addInt(splits.intdata[split]);
		array.addInt(splits.pointer);
	}

	public void add_split(){
		splits.addInt(array.pointer);
	}

	/**
	 * Convert the working arraylist to a primitive array.
	 */
	public int[] getPrimArray(int start){
		int[] primArray = new int[array.pointer-start];
		for(int i = 0; i < primArray.length; i++) 
			primArray[i] = array.intdata[i+start];
		return primArray;
	}

	/**
	 * Get the id of a string, given the current string mapping. If it is not encountered yet,
	 * then add it to the string mapping.
	 * @param s String to get the id from.
	 */
	public int strID(String s){ 
		Integer id = strStor.getInt(s);
		if(id==null) id = strStor.add(s);
		return id;
	} 

	/**
	 * Output function: convert int[] to String.
	 */
	public String arStr(int[] ar){
		if(ar==null) return "NULL";
		String s = "[";
		for(int i = 0; i < ar.length; i++){
			s+=ar[i];
			if(i!=ar.length-1)s+=",";
		}
		return s + "]";
	}

	////////////////////////
	/// DATATYPE STORAGE ///
	////////////////////////
	public int new_predicate(String name){
		array.addInt(IntHarvester.PREDICATE); 
		array.addInt(strID(name));
		array.addInt(0);
		return array.pointer-1;
	}

	public void replace(int index, int newint){
		array.intdata[index] = newint;
	}

	public void new_num(Token num, double multiplier){
		array.addInt(IntHarvester.NUMBER);
		double d = multiplier * Double.parseDouble(num.image);
		long l = Double.doubleToLongBits(d);
		int a = (int)(l>>>32);
		int b = (int)((l<<32)>>>32);
		array.addInt(a);
		array.addInt(b);
	}

	public void new_cut(){
		array.addInt(IntHarvester.CUT);
		array.addInt(0);
	}

	public int new_list(){
		array.addInt(IntHarvester.LIST);
		array.addInt(0);
		return array.pointer-1;
	}

	public void finish_list(int curSize, int nrArgs){
		array.intdata[curSize] = Math.abs(nrArgs); 
		if(nrArgs>0){
			array.addInt(IntHarvester.LIST);
			array.addInt(0);
		} 
	}

	public int new_parenthesized(){
		array.addInt(IntHarvester.PARENTHESIZED);
		array.addInt(0);
		return array.pointer-1;
	}

	public void handle_var(String v){
		array.addInt(IntHarvester.VARIABLE);
		if(v.equals("_"))v+="G"+vars.size(); // anonymous variables share names in notation
		if(vars.get(v)==null) vars.put(v,vars.size()); // add it if needed
		int id = vars.get(v);
		if(varIndices.get(id)==null)varIndices.put(id,new PrimIntArrayList(true));
		varIndices.get(id).addInt(array.pointer); // remember where the variable was found
		array.addInt(id);
	}
}
